import unittest

from datarobot.helpers import partitioning_methods as pm
from datarobot.helpers.partitioning_methods import BasePartitioningMethod


class TestPartitionMethodDispatch(unittest.TestCase):
    def test_random_cv(self):
        partition_data = dict(cv_method="random", validation_type="CV", holdout_pct=20, reps=5)
        part = BasePartitioningMethod.from_data(partition_data)
        self.assertIsInstance(part, pm.RandomCV)

    def test_random_tvh(self):
        partition_data = dict(
            cv_method="random", validation_type="TVH", holdout_pct=20, validation_pct=16
        )
        part = BasePartitioningMethod.from_data(partition_data)
        self.assertIsInstance(part, pm.RandomTVH)

    def test_stratified_cv(self):
        partition_data = dict(cv_method="stratified", validation_type="CV", holdout_pct=20, reps=5)
        part = BasePartitioningMethod.from_data(partition_data)
        self.assertIsInstance(part, pm.StratifiedCV)

    def test_stratified_tvh(self):
        partition_data = dict(
            cv_method="stratified", validation_type="TVH", holdout_pct=20, validation_pct=16
        )
        part = BasePartitioningMethod.from_data(partition_data)
        self.assertIsInstance(part, pm.StratifiedTVH)

    def test_user_cv(self):
        partition_data = dict(
            cv_method="user",
            validation_type="CV",
            user_partition_col="partitioner",
            cv_holdout_level="holdout_level",
        )
        part = BasePartitioningMethod.from_data(partition_data)
        self.assertIsInstance(part, pm.UserCV)

    def test_user_tvh(self):
        partition_data = dict(
            cv_method="user",
            validation_type="TVH",
            user_partition_col="partitioner",
            training_level="training",
            validation_level="validation",
            holdout_level="holdout",
        )
        part = BasePartitioningMethod.from_data(partition_data)
        self.assertIsInstance(part, pm.UserTVH)

    def test_group_cv(self):
        partition_data = dict(
            cv_method="group",
            validation_type="CV",
            partition_key_cols=["groupby1", "groupby2"],
            holdout_pct=20,
            reps=5,
        )
        part = BasePartitioningMethod.from_data(partition_data)
        self.assertIsInstance(part, pm.GroupCV)

    def test_group_tvh(self):
        partition_data = dict(
            cv_method="group",
            validation_type="TVH",
            partition_key_cols=["groupby1", "groupby2"],
            holdout_pct=20,
            validation_pct=16,
        )
        part = BasePartitioningMethod.from_data(partition_data)
        self.assertIsInstance(part, pm.GroupTVH)

    def test_you_screwed_up(self):
        with self.assertRaises(ValueError):
            BasePartitioningMethod.from_data({"bananas": "and cream"})

    def test_none_in_none_out(self):
        part = BasePartitioningMethod.from_data(None)
        self.assertIsNone(part)



from collections import OrderedDict
import json

import dateutil
import pandas as pd
import pytest
import responses

from datarobot import Deployment
from datarobot.enums import ACCURACY_METRIC
from datarobot.models.deployment import Accuracy, AccuracyOverTime
from datarobot.utils import from_api


@pytest.fixture
def accuracy_response_data(deployment_data):
    return {
        "model_id": deployment_data["model"]["id"],
        "period": {"start": "2019-08-01T00:00:00.000000Z", "end": "2019-08-10T00:00:00.000000Z"},
        "metrics": {
            "AUC": {"baselineValue": 1, "value": 1.5, "percentChange": 50},
            "LogLoss": {"baselineValue": None, "value": None, "percentChange": None},
        },
    }


@pytest.fixture
def deployment_get_accuracy_response(unittest_endpoint, deployment_data, accuracy_response_data):
    url = "{}/deployments/{}/accuracy/".format(unittest_endpoint, deployment_data["id"])
    responses.add(
        responses.GET,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(accuracy_response_data),
    )


@pytest.fixture
def accuracy_over_time_response_data(deployment_data):
    return {
        "model_id": deployment_data["model"]["id"],
        "metric": "LogLoss",
        "summary": {
            "period": {
                "start": "2019-08-01T00:00:00.000000Z",
                "end": "2019-08-04T00:00:00.000000Z",
            },
            "value": 1,
            "sampleSize": 100,
        },
        "baseline": {"period": None, "value": None, "sampleSize": 0},
        "buckets": [
            {
                "period": {
                    "start": "2019-08-01T00:00:00.000000Z",
                    "end": "2019-08-02T00:00:00.000000Z",
                },
                "value": 0.22,
                "sampleSize": 30,
            },
            {
                "period": {
                    "start": "2019-08-02T00:00:00.000000Z",
                    "end": "2019-08-03T00:00:00.000000Z",
                },
                "value": 0.42,
                "sampleSize": 70,
            },
            {
                "period": {
                    "start": "2019-08-03T00:00:00.000000Z",
                    "end": "2019-08-04T00:00:00.000000Z",
                },
                "value": None,
                "sampleSize": 0,
            },
        ],
    }


@pytest.fixture
def deployment_get_accuracy_over_time_response(
    unittest_endpoint, deployment_data, accuracy_over_time_response_data
):
    url = "{}/deployments/{}/accuracyOverTime/".format(unittest_endpoint, deployment_data["id"])
    responses.add(
        responses.GET,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(accuracy_over_time_response_data),
    )


def assert_bucket(actual, expected):
    assert sorted(actual.keys()) == sorted(expected.keys())

    actual_period = actual["period"]
    if actual_period:
        assert actual["period"].keys() == expected["period"].keys()
        assert actual_period["start"] == dateutil.parser.parse(expected["period"]["start"])
        assert actual_period["end"] == dateutil.parser.parse(expected["period"]["end"])
    else:
        assert actual_period is None

    assert actual["value"] == expected["value"]
    assert actual["sample_size"] == expected["sample_size"]


@responses.activate
@pytest.mark.usefixtures("deployment_get_response", "deployment_get_accuracy_response")
def test_accuracy(deployment_data, accuracy_response_data):
    accuracy_objs = [
        Accuracy.get(deployment_data["id"]),
        Deployment.get(deployment_data["id"]).get_accuracy(),
    ]
    for accuracy in accuracy_objs:
        assert accuracy.model_id == accuracy_response_data["model_id"]

        expected_start_date = dateutil.parser.parse(accuracy_response_data["period"]["start"])
        expected_end_date = dateutil.parser.parse(accuracy_response_data["period"]["end"])
        assert accuracy.period["start"] == expected_start_date
        assert accuracy.period["end"] == expected_end_date

        assert accuracy.metrics.keys() == accuracy_response_data["metrics"].keys()
        for metric_name, metric in accuracy.metrics.items():
            expected_metric = from_api(
                accuracy_response_data["metrics"][metric_name], keep_null_keys=True
            )
            assert metric == expected_metric
            assert accuracy[metric_name] == expected_metric["value"]

        expected_metric_values = {
            name: metric["value"] for name, metric in accuracy_response_data["metrics"].items()
        }
        assert accuracy.metric_values == expected_metric_values
        expected_metric_baseline = {
            name: metric["baselineValue"]
            for name, metric in accuracy_response_data["metrics"].items()
        }
        assert accuracy.metric_baselines == expected_metric_baseline
        expected_percent_changes = {
            name: metric["percentChange"]
            for name, metric in accuracy_response_data["metrics"].items()
        }
        assert accuracy.percent_changes == expected_percent_changes


@responses.activate
@pytest.mark.usefixtures("deployment_get_response", "deployment_get_accuracy_over_time_response")
def test_accuracy_over_time(deployment_data, accuracy_over_time_response_data):
    accuracy_objs = [
        AccuracyOverTime.get(deployment_data["id"]),
        Deployment.get(deployment_data["id"]).get_accuracy_over_time(),
    ]
    for accuracy_over_time in accuracy_objs:
        expected = from_api(accuracy_over_time_response_data, keep_null_keys=True)
        assert accuracy_over_time.metric == expected["metric"]
        assert accuracy_over_time.model_id == expected["model_id"]

        assert_bucket(accuracy_over_time.summary, expected["summary"])
        assert_bucket(accuracy_over_time.baseline, expected["baseline"])
        for index, bucket in enumerate(accuracy_over_time.buckets):
            assert_bucket(bucket, expected["buckets"][index])

        expected_bucket_values = [
            (dateutil.parser.parse(bucket["period"]["start"]), bucket["value"])
            for bucket in expected["buckets"]
        ]
        assert accuracy_over_time.bucket_values == OrderedDict(expected_bucket_values)

        expected_bucket_sample_sizes = [
            (dateutil.parser.parse(bucket["period"]["start"]), bucket["sample_size"])
            for bucket in expected["buckets"]
        ]
        assert accuracy_over_time.bucket_sample_sizes == OrderedDict(expected_bucket_sample_sizes)


@responses.activate
@pytest.mark.usefixtures("deployment_get_accuracy_over_time_response")
def test_accuracy_over_time_retrieve_as_dataframe(
    deployment_data, accuracy_over_time_response_data
):
    metric_names = [ACCURACY_METRIC.LOGLOSS, ACCURACY_METRIC.LOGLOSS]
    actual = AccuracyOverTime.get_as_dataframe(deployment_data["id"], metrics=metric_names)

    data = []
    start_times = []
    for bucket in accuracy_over_time_response_data["buckets"]:
        start_time = dateutil.parser.parse(bucket["period"]["start"])
        start_times.append(start_time)
        data.append([bucket["value"]] * 2)
    expected = pd.DataFrame(data, index=start_times, columns=metric_names)

    assert actual.equals(expected)



import trafaret as t

UserBlueprintsHexColumnNameLookupEntry_ = t.Dict(
    {
        t.Key("colname"): t.String(allow_blank=False),
        t.Key("hex"): t.String(allow_blank=False),
        t.Key("project_id", optional=True): t.String(allow_blank=False),
    }
)

ParamValuePair_ = t.Dict(
    {
        t.Key("param"): t.String(allow_blank=False),
        t.Key("value", optional=True): t.Or(
            t.Or(t.Int(), t.String(allow_blank=True), t.Bool(), t.Float(), t.Null()),
            t.List(t.Or(t.Int(), t.String(allow_blank=True), t.Bool(), t.Float(), t.Null())),
        ),
    }
)

UserBlueprintsBlueprintTaskData_ = t.Dict(
    {
        t.Key("inputs"): t.List(t.String(allow_blank=False)),
        t.Key("output_method"): t.String(allow_blank=False),
        t.Key("output_method_parameters"): t.List(ParamValuePair_),
        t.Key("task_code"): t.String(allow_blank=False),
        t.Key("task_parameters"): t.List(ParamValuePair_),
        t.Key("x_transformations"): t.List(ParamValuePair_),
        t.Key("y_transformations"): t.List(ParamValuePair_),
    }
)

UserBlueprintsBlueprintTask_ = t.Dict(
    {
        t.Key("task_id"): t.String(allow_blank=False),
        t.Key("task_data"): UserBlueprintsBlueprintTaskData_,
    }
)

VertexContextItemInfo_ = t.Dict(
    {
        t.Key("inputs"): t.List(t.String(allow_blank=False)),
        t.Key("outputs"): t.List(t.String(allow_blank=False)),
    }
)

VertexContextItemMessages_ = t.Dict(
    {
        t.Key("errors", optional=True): t.List(t.String(allow_blank=False)),
        t.Key("warnings", optional=True): t.List(t.String(allow_blank=False)),
    }
)

VertexContextItem_ = t.Dict(
    {
        t.Key("task_id"): t.String(allow_blank=False),
        t.Key("information"): t.Or(VertexContextItemInfo_),
        t.Key("messages"): t.Or(VertexContextItemMessages_),
    }
)

UserBlueprint_ = t.Dict(
    {
        t.Key("blender"): t.Bool(),
        t.Key("blueprint_id"): t.String(allow_blank=False),
        t.Key("custom_task_version_metadata", optional=True): t.List(
            t.List(t.String(allow_blank=False))
        ),
        t.Key("diagram"): t.String(allow_blank=False),
        t.Key("features"): t.List(t.String(allow_blank=False)),
        t.Key("features_text"): t.String(allow_blank=True),
        t.Key("hex_column_name_lookup", optional=True): t.List(
            UserBlueprintsHexColumnNameLookupEntry_
        ),
        t.Key("icons"): t.List(t.Int()),
        t.Key("insights"): t.String(allow_blank=False),
        t.Key("is_time_series", default=False): t.Bool(),
        t.Key("model_type"): t.String(allow_blank=False),
        t.Key("project_id", optional=True): t.String(allow_blank=False),
        t.Key("reference_model", default=False): t.Bool(),
        t.Key("shap_support", default=False): t.Bool(),
        t.Key("supported_target_types"): t.List(t.String(allow_blank=False)),
        t.Key("supports_gpu", default=False): t.Bool(),
        t.Key("user_blueprint_id"): t.String(allow_blank=False),
        t.Key("user_id"): t.String(allow_blank=False),
        t.Key("blueprint", optional=True): t.List(UserBlueprintsBlueprintTask_),
        t.Key("vertex_context", optional=True): t.List(VertexContextItem_),
    }
).allow_extra("*")


UserBlueprintsInputType_ = t.Dict(
    {t.Key("type"): t.String(allow_blank=False), t.Key("name"): t.String(allow_blank=False)}
)

UserBlueprintsInputTypesResponse_ = t.Dict({t.Key("input_types"): t.List(UserBlueprintsInputType_)})


UserBlueprintAddedToMenuItem_ = t.Dict(
    {
        t.Key("blueprint_id"): t.String(allow_blank=False),
        t.Key("user_blueprint_id"): t.String(allow_blank=False),
    }
)

UserBlueprintAddToMenuResponse_ = t.Dict(
    {t.Key("added_to_menu"): t.List(UserBlueprintAddedToMenuItem_)}
)


UserBlueprintsValidateTaskParameter_ = t.Dict(
    {
        t.Key("message"): t.String(allow_blank=False),
        t.Key("param_name"): t.String(allow_blank=False),
        t.Key("value", optional=True): t.Or(
            t.Or(t.Int(), t.String(allow_blank=True), t.Bool(), t.Float(), t.Null()),
            t.List(t.Or(t.Int(), t.String(allow_blank=True), t.Bool(), t.Float(), t.Null())),
        ),
    }
)

UserBlueprintsValidateTaskParametersResponse_ = t.Dict(
    {t.Key("errors"): t.List(UserBlueprintsValidateTaskParameter_)}
)


UserBlueprintTaskCategoryItem_ = t.Dict(
    {
        t.Key("name"): t.String(allow_blank=False),
        t.Key("task_codes"): t.List(t.String(allow_blank=False)),
        t.Key("subcategories", optional=True): t.List(t.Dict().allow_extra("*")),
    }
).allow_extra("*")

UserBlueprintTaskArgumentDefinition_ = t.Dict(
    {
        t.Key("name"): t.String(allow_blank=False),
        t.Key("type"): t.String(allow_blank=False),
        t.Key("default", optional=True): t.Or(
            t.Or(t.Int(), t.String(allow_blank=True), t.Bool(), t.Float(), t.Null()),
            t.List(t.Or(t.Int(), t.String(allow_blank=True), t.Bool(), t.Float(), t.Null())),
        ),
        t.Key("values"): t.Or(
            t.Or(t.Int(), t.String(allow_blank=True), t.Bool(), t.Float(), t.Null()),
            t.List(t.Or(t.Int(), t.String(allow_blank=True), t.Bool(), t.Float(), t.Null())),
            t.Dict().allow_extra("*"),
        ),
        t.Key("tunable", optional=True): t.Bool(),
        t.Key("recommended", optional=True): t.Or(
            t.Or(t.Int(), t.String(allow_blank=True), t.Bool(), t.Float(), t.Null()),
            t.List(t.Or(t.Int(), t.String(allow_blank=True), t.Bool(), t.Float(), t.Null())),
        ),
    }
).allow_extra("*")

UserBlueprintTaskArgument_ = t.Dict(
    {
        t.Key("key"): t.String(allow_blank=False),
        t.Key("argument"): t.Or(UserBlueprintTaskArgumentDefinition_),
    }
)

ColnameAndType_ = t.Dict(
    {
        t.Key("hex"): t.String(allow_blank=False),
        t.Key("colname"): t.String(allow_blank=False),
        t.Key("type"): t.String(allow_blank=False),
    }
)

TaskDocumentationUrl_ = t.Dict(
    {t.Key("documentation", optional=True): t.String(allow_blank=True)}
).allow_extra("*")

UserBlueprintTaskCustomTaskMetadata_ = t.Dict(
    {
        t.Key("id"): t.String(allow_blank=False),
        t.Key("version_major"): t.Int(),
        t.Key("version_minor"): t.Int(),
        t.Key("label"): t.String(allow_blank=False),
    }
)

UserBlueprintTask_ = t.Dict(
    {
        t.Key("task_code"): t.String(allow_blank=False),
        t.Key("label"): t.String(allow_blank=True),
        t.Key("description"): t.String(allow_blank=True),
        t.Key("arguments"): t.List(UserBlueprintTaskArgument_),
        t.Key("categories"): t.List(t.String(allow_blank=False)),
        t.Key("colnames_and_types", optional=True): t.Or(t.List(ColnameAndType_), t.Null),
        t.Key("icon"): t.Int(),
        t.Key("output_methods"): t.List(t.String(allow_blank=False)),
        t.Key("time_series_only"): t.Bool(),
        t.Key("url"): t.Or(
            t.Dict().allow_extra("*"), t.String(allow_blank=True), TaskDocumentationUrl_,
        ),
        t.Key("valid_inputs"): t.List(t.String(allow_blank=False)),
        t.Key("is_custom_task", optional=True): t.Bool(),
        t.Key("custom_task_id", optional=True): t.Or(t.String, t.Null),
        t.Key("custom_task_versions", optional=True): t.List(UserBlueprintTaskCustomTaskMetadata_),
        t.Key("supports_scoring_code", optional=True): t.Bool(),
    }
).allow_extra("*")

UserBlueprintTaskLookupEntry_ = t.Dict(
    {
        t.Key("task_code"): t.String(allow_blank=False),
        t.Key("task_definition"): t.Or(UserBlueprintTask_),
    }
)

UserBlueprintTasksResponse_ = t.Dict(
    {
        t.Key("categories"): t.List(UserBlueprintTaskCategoryItem_),
        t.Key("tasks"): t.List(UserBlueprintTaskLookupEntry_),
    }
)

UserBlueprintSharedRolesResponseValidator_ = t.Dict(
    {
        t.Key("share_recipient_type"): t.Enum("user", "group", "organization"),
        t.Key("role"): t.Enum("CONSUMER", "EDITOR", "OWNER"),
        t.Key("id"): t.String(allow_blank=False, min_length=24, max_length=24),
        t.Key("name"): t.String(allow_blank=False),
    }
)

UserBlueprintSharedRolesListResponseValidator_ = t.Dict(
    {
        t.Key("count", optional=True): t.Int(),
        t.Key("next", optional=True): t.URL,
        t.Key("previous", optional=True): t.URL,
        t.Key("data"): t.List(UserBlueprintSharedRolesResponseValidator_),
        t.Key("total_count", optional=True): t.Int(),
    }
)


UserBlueprintCatalogSearchItem_ = t.Dict(
    {
        t.Key("id"): t.String(),
        t.Key("catalog_name"): t.String(),
        t.Key("description", optional=True): t.String(),
        t.Key("info_creator_full_name"): t.String(),
        t.Key("last_modifier_name", optional=True): t.String(),
        t.Key("user_blueprint_id"): t.String(),
    }
).allow_extra("*")



from collections import defaultdict

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2


class WordCloud(APIObject):
    """ Word cloud data for the model.

    Notes
    ----------
    ``WordCloudNgram`` is a dict containing the following:

        * ``ngram`` (str) Word or ngram value.
        * ``coefficient`` (float) Value from [-1.0, 1.0] range, describes effect of this ngram on \
          the target. Large negative value means strong effect toward negative class in \
          classification and smaller target value in regression models. Large positive - toward \
          positive class and bigger value respectively.
        * ``count`` (int) Number of rows in the training sample where this ngram appears.
        * ``frequency`` (float) Value from (0.0, 1.0] range, relative frequency of given ngram to \
          most frequent ngram.
        * ``is_stopword`` (bool) True for ngrams that DataRobot evaluates as stopwords.
        * ``class`` (str or None) For classification - values of the target class for
          corresponding word or ngram. For regression - None.

    Attributes
    ----------
    ngrams : list of dicts
        List of dicts with schema described as ``WordCloudNgram`` above.
    """

    _converter = t.Dict(
        {
            t.Key("ngrams"): t.List(
                t.Dict(
                    {
                        t.Key("ngram"): t.String,
                        t.Key("coefficient"): t.Float(gte=-1, lte=1),
                        t.Key("count"): t.Int,
                        t.Key("frequency"): t.Float(gt=0, lte=1),
                        t.Key("is_stopword"): t.Bool,
                        # Making these optional will allow working with older backends
                        t.Key("class", optional=True, default=None): t.String | t.Null,
                        t.Key("variable", optional=True, default=None): t.String | t.Null,
                    }
                ).ignore_extra("*")
            )
        }
    ).ignore_extra("*")

    def __init__(self, ngrams):
        self.ngrams = ngrams

    def __repr__(self):
        return encode_utf8_if_py2(u"WordCloud({} ngrams)".format(len(self.ngrams)))

    def most_frequent(self, top_n=5):
        """ Return most frequent ngrams in the word cloud.

        Parameters
        ----------
        top_n : int
            Number of ngrams to return

        Returns
        -------
        list of dict
            Up to top_n top most frequent ngrams in the word cloud.
            If top_n bigger then total number of ngrams in word cloud - return all sorted by
            frequency in descending order.
        """
        return sorted(self.ngrams, key=lambda ngram: ngram["frequency"], reverse=True)[:top_n]

    def most_important(self, top_n=5):
        """ Return most important ngrams in the word cloud.

        Parameters
        ----------
        top_n : int
            Number of ngrams to return

        Returns
        -------
        list of dict
            Up to top_n top most important ngrams in the word cloud.
            If top_n bigger then total number of ngrams in word cloud - return all sorted by
            absolute coefficient value in descending order.
        """
        return sorted(self.ngrams, key=lambda ngram: abs(ngram["coefficient"]), reverse=True)[
            :top_n
        ]

    def ngrams_per_class(self):
        """ Split ngrams per target class values. Useful for multiclass models.

        Returns
        -------
        dict
            Dictionary in the format of (class label) -> (list of ngrams for that class)
        """
        per_class = defaultdict(list)
        for ngram in self.ngrams:
            per_class[ngram["class"]].append(ngram)
        return dict(per_class)



from mock import patch
import pandas
import pytest
import responses

from datarobot import PredictionExplanations, PredictionExplanationsInitialization
from datarobot.models.prediction_explanations import (
    PredictionExplanationsPage,
    PredictionExplanationsRow,
)


@pytest.fixture
def pei_url(project_collection_url, pei_data):
    return "{}{}/models/{}/predictionExplanationsInitialization/".format(
        project_collection_url, pei_data["project_id"], pei_data["model_id"]
    )


def test_prediction_explanations_initialization_future_proof(pei_data):
    PredictionExplanationsInitialization.from_data(dict(pei_data, future="new"))


def test_prediction_explanations_future_proof(pe_data):
    PredictionExplanations.from_data(dict(pe_data, future="new"))


def test_prediction_explanations_page_future_proof(pep_data):
    with patch(
        "datarobot.models.prediction_explanations.t.URL.check",
        side_effect=(lambda url, context=None: url),
    ):
        PredictionExplanationsPage.from_data(dict(pep_data, future="new"))


@responses.activate
def test_prediction_explanations_initialization_get_returns_valid_data(pei_url, pei_json, one_pei):
    responses.add(
        responses.GET, pei_url, body=pei_json, status=200, content_type="application/json"
    )
    initialization = PredictionExplanationsInitialization.get(one_pei.project_id, one_pei.model_id)
    assert isinstance(initialization, PredictionExplanationsInitialization)
    assert initialization.project_id == one_pei.project_id
    assert initialization.model_id == one_pei.model_id
    assert initialization.prediction_explanations_sample


@responses.activate
def test_prediction_explanations_initialization_create(
    pei_data, pei_url, project_url, base_job_server_data
):
    job_url = "{}jobs/{}/".format(project_url, base_job_server_data["id"])
    responses.add(
        responses.POST,
        pei_url,
        body="",
        status=202,
        content_type="application/json",
        adding_headers={"Location": job_url},
    )
    responses.add(
        responses.GET,
        job_url,
        json=base_job_server_data,
        status=303,
        adding_headers={"Location": pei_url},
    )

    pei_job = PredictionExplanationsInitialization.create(
        project_id=pei_data["project_id"], model_id=pei_data["model_id"]
    )
    assert str(pei_job.id) == base_job_server_data["id"]


@responses.activate
def test_prediction_explanations_initialization_delete(pei_url, pei_data):
    responses.add(responses.DELETE, pei_url, status=200)

    pei = PredictionExplanationsInitialization.from_data(pei_data)
    pei.delete()


@responses.activate
def test_prediction_explanations_get_returns_valid_data(project_id, project_url, pe_json, one_pe):
    url = "{}predictionExplanationsRecords/{}/".format(project_url, one_pe.id)
    responses.add(responses.GET, url, body=pe_json, status=200, content_type="application/json")
    record = PredictionExplanations.get(project_id, one_pe.id)
    assert isinstance(record, PredictionExplanations)
    assert record.id == one_pe.id
    assert record.model_id == one_pe.model_id


@responses.activate
def test_prediction_explanations_page_get_returns_valid_data(
    project_id, project_url, pep_json, one_pep
):
    url = "{}predictionExplanations/{}/".format(project_url, one_pep.id)
    responses.add(responses.GET, url, body=pep_json, status=200, content_type="application/json")
    with patch(
        "datarobot.models.prediction_explanations.t.URL.check",
        side_effect=(lambda url, context=None: url),
    ):
        record = PredictionExplanationsPage.get(project_id, one_pep.id)
    assert isinstance(record, PredictionExplanationsPage)
    assert record.id == one_pep.id
    assert record.data == one_pep.data
    assert record.adjustment_method == "N/A"


@responses.activate
def test_prediction_explanations_page_get_returns_valid_data_adjusted(
    project_id, project_url, pep_data_with_adjusted_predictions, one_pep
):
    url = "{}predictionExplanations/{}/".format(project_url, one_pep.id)
    responses.add(
        responses.GET,
        url,
        json=pep_data_with_adjusted_predictions,
        status=200,
        content_type="application/json",
    )
    with patch(
        "datarobot.models.prediction_explanations.t.URL.check",
        side_effect=(lambda url, context=None: url),
    ):
        record = PredictionExplanationsPage.get(project_id, one_pep.id)
    assert isinstance(record, PredictionExplanationsPage)
    assert record.id == one_pep.id
    assert record.data == one_pep.data
    assert record.adjustment_method == "exposureNormalized"
    for pe_row in record.data:
        assert pe_row["adjusted_prediction"] == pe_row["prediction"]
        assert pe_row["adjusted_prediction_values"] == pe_row["prediction_values"]


@responses.activate
def test_prediction_explanations_create(
    project_id, project_url, project_collection_url, base_job_server_data, pep_data
):
    pep_base_url = "{}{}/predictionExplanations/".format(project_collection_url, project_id)
    job_url = "{}jobs/{}/".format(project_url, base_job_server_data["id"])
    pep_url = "{}/{}/".format(pep_base_url, pep_data["id"])

    responses.add(
        responses.POST,
        pep_base_url,
        body="",
        status=202,
        content_type="application/json",
        adding_headers={"Location": job_url},
    )
    responses.add(
        responses.GET,
        job_url,
        json=base_job_server_data,
        status=303,
        adding_headers={"Location": pep_url},
    )

    pep_job = PredictionExplanations.create(
        project_id=project_id,
        model_id="578e59a41ced2e5a9eb18960",
        dataset_id="581c6256100d2b60586980d9",
    )
    assert str(pep_job.id) == base_job_server_data["id"]


@responses.activate
def test_prediction_explanations_list(project_id, project_url, pe_list_json):
    pe_base_url = "{}predictionExplanationsRecords/".format(project_url)
    responses.add(
        responses.GET,
        pe_base_url + "?modelId=578e59a41ced2e5a9eb18965",
        status=200,
        body=pe_list_json,
        content_type="application/json",
        match_querystring=True,
    )
    pe_list = PredictionExplanations.list(project_id, model_id="578e59a41ced2e5a9eb18965")
    assert len(pe_list) == 3
    for pe in pe_list:
        assert isinstance(pe, PredictionExplanations)


@responses.activate
def test_prediction_explanations_delete(project_collection_url, project_id, pe_data):
    pe_url = "{}{}/predictionExplanationsRecords/{}/".format(
        project_collection_url, project_id, pe_data["id"]
    )
    responses.add(responses.DELETE, pe_url, status=200)

    pe = PredictionExplanations.from_data(pe_data)
    pe.delete()


@responses.activate
def test_prediction_explanations_iteration(project_id, project_url, one_pep):
    url = "{}predictionExplanations/{}/".format(project_url, one_pep.id)
    url_page1 = url + "?offset=0&excludeAdjustedPredictions=true"
    url_page2 = url + "?limit=2&offset=2&excludeAdjustedPredictions=true"
    url_page3 = url + "?limit=2&offset=4&excludeAdjustedPredictions=true"
    resp_data_1 = {
        "id": one_pep.id,
        "count": 5,
        "previous": None,
        "next": url_page2,
        "data": one_pep.data[:2],
        "prediction_explanations_record_location": one_pep.prediction_explanations_record_location,
    }
    resp_data_2 = dict(resp_data_1, previous=url_page1, next=url_page3, data=one_pep.data[2:4])
    resp_data_3 = dict(resp_data_1, previous=url_page2, next=None, data=one_pep.data[4:5])
    responses.add(responses.GET, url_page1, json=resp_data_1, status=200, match_querystring=True)
    responses.add(responses.GET, url_page2, json=resp_data_2, status=200, match_querystring=True)
    responses.add(responses.GET, url_page3, json=resp_data_3, status=200, match_querystring=True)
    with patch(
        "datarobot.models.prediction_explanations.t.URL.check",
        side_effect=(lambda url, context=None: url),
    ):
        prediction_explanations_rows = list(
            PredictionExplanations(one_pep.id, project_id, "", "", "", 0, 0, "").get_rows()
        )
    assert len(prediction_explanations_rows) == 5
    for i, row in enumerate(prediction_explanations_rows):
        assert isinstance(row, PredictionExplanationsRow)
        assert row.row_id == one_pep.data[i]["row_id"]
        assert row.prediction == one_pep.data[i]["prediction"]
        assert row.prediction_values == one_pep.data[i]["prediction_values"]
        assert row.prediction_explanations == one_pep.data[i]["prediction_explanations"]


@responses.activate
def test_prediction_explanations_as_dataframe_regression(project_url, pep_data, one_pe):
    url = "{}predictionExplanations/{}/".format(project_url, one_pe.id)
    responses.add(responses.GET, url, json=pep_data, status=200)
    with patch(
        "datarobot.models.prediction_explanations.t.URL.check",
        side_effect=(lambda url, context=None: url),
    ):
        prediction_explanations_as_dataframe = one_pe.get_all_as_dataframe()
    assert isinstance(prediction_explanations_as_dataframe, pandas.DataFrame)
    rows = pep_data["count"]
    # 5 columns per prediction explanation, plus prediction and row id
    columns = 5 * one_pe.max_explanations + 2
    assert prediction_explanations_as_dataframe.shape == (rows, columns)


@responses.activate
def test_prediction_explanations_as_dataframe_regression_with_exposure(
    project_url, one_pe, pep_data_with_adjusted_predictions
):
    url = "{}predictionExplanations/{}/".format(project_url, one_pe.id)
    responses.add(responses.GET, url, json=pep_data_with_adjusted_predictions, status=200)
    with patch(
        "datarobot.models.prediction_explanations.t.URL.check",
        side_effect=(lambda url, context=None: url),
    ):
        prediction_explanations_as_dataframe = one_pe.get_all_as_dataframe()
    assert isinstance(prediction_explanations_as_dataframe, pandas.DataFrame)
    rows = pep_data_with_adjusted_predictions["count"]
    # 5 columns per prediction explanation, plus prediction, adjusted prediction and row id
    columns = 5 * one_pe.max_explanations + 3
    assert prediction_explanations_as_dataframe.shape == (rows, columns)


@responses.activate
def test_prediction_explanations_as_dataframe_classification(
    project_url, pep_data_classification, one_pe
):
    url = "{}predictionExplanations/{}/".format(project_url, one_pe.id)
    responses.add(responses.GET, url, json=pep_data_classification, status=200)
    with patch(
        "datarobot.models.prediction_explanations.t.URL.check",
        side_effect=(lambda url, context=None: url),
    ):
        prediction_explanations_as_dataframe = one_pe.get_all_as_dataframe()
    assert isinstance(prediction_explanations_as_dataframe, pandas.DataFrame)
    rows = pep_data_classification["count"]
    # 5 columns per explanation, plus 2 columns per target level, plus row_id and prediction
    columns = (5 * one_pe.max_explanations) + (2 * 2) + 2
    assert prediction_explanations_as_dataframe.shape == (rows, columns)
    for idx, row in prediction_explanations_as_dataframe.iterrows():
        assert "class_0_label" in row
        assert row["class_0_label"] == 0


def test_prediction_explanations_download_to_csv():
    patch_target = (
        "datarobot.models.prediction_explanations.PredictionExplanations.get_all_as_dataframe"
    )
    with patch(patch_target) as pemock:
        prediction_explanations = PredictionExplanations(
            "fake_id", "fake_project_id", "", "", 3, 0, 0, ""
        )
        prediction_explanations.download_to_csv("test_file.csv", encoding="ascii")
        pemock.return_value.to_csv.assert_called_once_with(
            path_or_buf="test_file.csv", header=True, index=False, encoding="ascii"
        )



import trafaret as t

from datarobot.enums import DEFAULT_MAX_WAIT
from datarobot.models.api_object import APIObject
from datarobot.models.validators import feature_impact_trafaret
from datarobot.utils import deprecated, encode_utf8_if_py2, logger
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution

from .job import filter_feature_impact_result

logger = logger.get_logger(__name__)


class CustomInferenceImage(APIObject):
    """An image of a custom model.

    .. versionadded:: v2.21
    .. deprecated:: v2.23

    Attributes
    ----------
    id: str
        image id
    custom_model: dict
        dict with 2 keys: `id` and `name`,
        where `id` is the ID of the custom model
        and `name` is the model name
    custom_model_version: dict
        dict with 2 keys: `id` and `label`,
        where `id` is the ID of the custom model version
        and `label` is the version label
    execution_environment: dict
        dict with 2 keys: `id` and `name`,
        where `id` is the ID of the execution environment
        and `name` is the environment name
    execution_environment_version: dict
        dict with 2 keys: `id` and `label`,
        where `id` is the ID of the execution environment version
        and `label` is the version label
    latest_test: dict, optional
        dict with 3 keys: `id`, `status` and `completedAt`,
        where `id` is the ID of the latest test,
        `status` is the testing status
        and `completedAt` is ISO-8601 formatted timestamp of when the testing was completed
    """

    _path = "customInferenceImages/"
    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("custom_model"): t.Dict({t.Key("id"): t.String(), t.Key("name"): t.String()}),
            t.Key("custom_model_version"): t.Dict(
                {t.Key("id"): t.String(), t.Key("label"): t.String()}
            ),
            t.Key("execution_environment"): t.Dict(
                {t.Key("id"): t.String(), t.Key("name"): t.String()}
            ),
            t.Key("execution_environment_version"): t.Dict(
                {t.Key("id"): t.String(), t.Key("label"): t.String()}
            ),
            t.Key("latest_test", optional=True): t.Dict(
                {
                    t.Key("id"): t.String(),
                    t.Key("status"): t.String(),
                    t.Key("completed_at"): t.String(allow_blank=True),
                }
            ),
        }
    ).ignore_extra("*")

    def __init__(self, **kwargs):
        self._set_values(**kwargs)

    def __repr__(self):
        return encode_utf8_if_py2(u"{}({!r})".format(self.__class__.__name__, self.id))

    def _set_values(
        self,
        id,
        custom_model,
        custom_model_version,
        execution_environment,
        execution_environment_version,
        latest_test=None,
    ):
        self.id = id
        self.custom_model = custom_model
        self.custom_model_version = custom_model_version
        self.execution_environment = execution_environment
        self.execution_environment_version = execution_environment_version
        self.latest_test = latest_test

    @classmethod
    @deprecated(
        deprecated_since_version="v2.23",
        will_remove_version="v2.24",
        message="CustomModelImages have been deprecated. "
        "Please use CustomModelVersions with base_environment_id",
    )
    def create(
        cls, custom_model_id, custom_model_version_id, environment_id, environment_version_id=None,
    ):
        """Create a custom model image.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            the id of the custom model
        custom_model_version_id: str
            the id of the custom model version
        environment_id: str
            the id of the execution environment
        environment_version_id: str, optional
            the id of the execution environment version

        Returns
        -------
        CustomInferenceImage
            created custom model image

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        payload = {
            "custom_model_id": custom_model_id,
            "custom_model_version_id": custom_model_version_id,
            "environment_id": environment_id,
            "environment_version_id": environment_version_id,
        }
        response = cls._client.post(cls._path, data=payload)
        return cls.from_server_data(response.json())

    @classmethod
    @deprecated(
        deprecated_since_version="v2.23",
        will_remove_version="v2.24",
        message="CustomModelImages have been deprecated. "
        "Please use CustomModelVersions with base_environment_id",
    )
    def list(
        cls,
        testing_status=None,
        custom_model_id=None,
        custom_model_version_id=None,
        environment_id=None,
        environment_version_id=None,
    ):
        """List custom model images.

        .. versionadded:: v2.21

        Parameters
        ----------
        testing_status: str, optional
            the testing status to filter results by
        custom_model_id: str, optional
            the id of the custom model
        custom_model_version_id: str, optional
            the id of the custom model version
        environment_id: str, optional
            the id of the execution environment
        environment_version_id: str, optional
            the id of the execution environment version

        Returns
        -------
        List[CustomModelImage]
            a list of custom model images

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        payload = {
            "testing_status": testing_status,
            "custom_model_id": custom_model_id,
            "custom_model_version_id": custom_model_version_id,
            "environment_id": environment_id,
            "environment_version_id": environment_version_id,
        }
        data = unpaginate(cls._path, payload, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    @deprecated(
        deprecated_since_version="v2.23",
        will_remove_version="v2.24",
        message="CustomModelImages have been deprecated. "
        "Please use CustomModelVersions with base_environment_id",
    )
    def get(cls, custom_model_image_id):
        """Get custom model image by id.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_image_id: str
            the id of the custom model image

        Returns
        -------
        CustomInferenceImage
            retrieved custom model image

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = "{}{}/".format(cls._path, custom_model_image_id)
        return cls.from_location(path)

    @deprecated(
        deprecated_since_version="v2.23",
        will_remove_version="v2.24",
        message="CustomModelImages have been deprecated. "
        "Please use CustomModelVersions with base_environment_id",
    )
    def refresh(self):
        """Update custom inference image with the latest data from server.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        path = "{}{}/".format(self._path, self.id)

        response = self._client.get(path)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))

    @deprecated(
        deprecated_since_version="v2.23",
        will_remove_version="v2.24",
        message="CustomModelImages have been deprecated. "
        "Please use CustomModelVersions with base_environment_id",
    )
    def get_feature_impact(self, with_metadata=False):
        """Get custom model feature impact.

        .. versionadded:: v2.21

        Parameters
        ----------
        with_metadata : bool
            The flag indicating if the result should include the metadata as well.

        Returns
        -------
        feature_impacts : list of dict
           The feature impact data. Each item is a dict with the keys 'featureName',
           'impactNormalized', and 'impactUnnormalized', and 'redundantWith'.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = "{}{}/featureImpact/".format(self._path, self.id)
        data = self._client.get(path).json()
        data = feature_impact_trafaret.check(data)
        return filter_feature_impact_result(data, with_metadata=with_metadata)

    @deprecated(
        deprecated_since_version="v2.23",
        will_remove_version="v2.24",
        message="CustomModelImages have been deprecated. "
        "Please use CustomModelVersions with base_environment_id",
    )
    def calculate_feature_impact(self, max_wait=DEFAULT_MAX_WAIT):
        """Calculate custom model feature impact.

        .. versionadded:: v2.22

        Parameters
        ----------
        max_wait: int, optional
            max time to wait for feature impact calculation.
            If set to None - method will return without waiting.
            Defaults to 10 min

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        path = "{}{}/featureImpact/".format(self._path, self.id)
        response = self._client.post(path)

        if max_wait is not None:
            wait_for_async_resolution(self._client, response.headers["Location"], max_wait)



import json

import pytest


@pytest.fixture
def async_url():
    return "http://host_name.com/status/status_id/"


@pytest.fixture
def resolved_url():
    return "https://host_name.com/finished/id/"


@pytest.fixture
def async_success_callback(resolved_url):
    def callback(request):
        return 303, {"Location": resolved_url}, ""

    return callback


@pytest.fixture
def async_pending_callback():
    def callback(request):
        return (
            200,
            {},
            json.dumps(
                dict(
                    status="INITIALIZED", message="", code=0, created="2016-07-22T12:00:00.123456Z"
                )
            ),
        )

    return callback


@pytest.fixture
def async_job_failure_callback():
    def callback(request):
        return (
            200,
            {},
            json.dumps(
                dict(status="ERROR", message="", code=0, created="2016-07-22T12:00:00.123456Z")
            ),
        )

    return callback


@pytest.fixture
def async_job_completed_callback():
    def callback(request):
        return (
            200,
            {},
            json.dumps(
                dict(status="COMPLETED", message="", code=0, created="2016-07-22T12:00:00.123456Z")
            ),
        )

    return callback


@pytest.fixture
def async_server_nonsense_callback():
    """
    A callback that returns a nonsensical status code.
    400-level and 500-level are caught and raised automatically by the client, so
    we'll have this callback return a 299 (that is not a real status code)
    to stand in for the "I don't know what the server is trying to say"
    """

    def callback(request):
        return 299, {}, json.dumps({"message": "An unexpected error has occurred"})

    return callback


@pytest.fixture
def async_not_finished_callback():
    def callback(request):
        return (
            200,
            {},
            json.dumps(
                dict(status="RUNNING", message="", code=0, created="2016-07-22T12:00:00.123456Z")
            ),
        )

    return callback


@pytest.fixture
def custom_resolver():
    def resolver(response):
        data = response.json()
        if data["status"] in ["COMPLETED", "ERROR"]:
            return data

    return resolver



from mock import patch
import pytest
import responses

from datarobot.errors import ClientError
from datarobot.models import ExternalConfusionChart
from datarobot.utils import from_api


class TestExternalConfusionChart(object):
    @pytest.fixture
    def confusion_chart_classes(self):
        return

    @staticmethod
    def create_response_data(confusion_chart_data, count):
        confusion_chart_data["classes"] = ["1", "2", "3"]
        return [
            {"data": confusion_chart_data, "datasetId": "dataset_id_{}".format(i)}
            for i in range(count)
        ]

    @pytest.fixture
    def url(self, project_id, model_id):
        return "https://host_name.com/projects/{}/models/{}/datasetConfusionCharts/".format(
            project_id, model_id
        )

    @pytest.fixture
    def metadata_url(self, project_id, model_id, dataset_id):
        return (
            "https://host_name.com/projects/{}/models/{}/datasetConfusionCharts/{}/metadata"
        ).format(project_id, model_id, dataset_id)

    @staticmethod
    def assert_equal_confusion_chart(confusion_chart_obj, api_chart_dict):
        one_chart_data = from_api(api_chart_dict["data"])
        assert confusion_chart_obj.dataset_id == api_chart_dict["datasetId"]
        assert confusion_chart_obj.class_metrics == one_chart_data["class_metrics"]
        assert confusion_chart_obj.confusion_matrix == one_chart_data["confusion_matrix"]
        assert confusion_chart_obj.classes == one_chart_data["classes"]

    def test_instantiation(self, confusion_chart_data):
        response_chart = self.create_response_data(confusion_chart_data, 1)[0]
        chart_obj = ExternalConfusionChart.from_server_data(response_chart)
        self.assert_equal_confusion_chart(chart_obj, response_chart)

    def test_future_proof(self, confusion_chart_data):
        response_chart = self.create_response_data(confusion_chart_data, 1)[0]
        data_with_future_keys = dict(response_chart, new_key="some future lift data")
        data_with_future_keys["data"]["new_key"] = {"new_key": "future class metrics"}
        ExternalConfusionChart.from_server_data(data_with_future_keys)

    @responses.activate
    @patch(
        "datarobot.models.external_dataset_scores_insights."
        "external_confusion_chart.DEFAULT_BATCH_SIZE",
        10,
    )
    def test_list_confusion_chart_all_results(
        self, confusion_chart_data, url, project_id, model_id
    ):
        url_page1 = url + "?limit=10&offset=0"
        url_page2 = url + "?limit=10&offset=10"
        all_data = self.create_response_data(confusion_chart_data, 11)
        resp_data_1 = dict(count=10, previous=None, next=url_page2, data=all_data[:10])
        resp_data_2 = dict(count=1, previous=url_page1, next=None, data=all_data[10:])
        responses.add(
            responses.GET, url_page1, json=resp_data_1, status=200, match_querystring=True
        )
        responses.add(
            responses.GET, url_page2, json=resp_data_2, status=200, match_querystring=True
        )

        confusion_charts = ExternalConfusionChart.list(project_id, model_id, limit=0)
        assert len(confusion_charts) == 11
        for i, chart in enumerate(confusion_charts):
            assert isinstance(chart, ExternalConfusionChart)
            self.assert_equal_confusion_chart(chart, all_data[i])

    @responses.activate
    @patch(
        "datarobot.models.external_dataset_scores_insights."
        "external_confusion_chart.DEFAULT_BATCH_SIZE",
        10,
    )
    def test_list_confusion_chart_all_results_with_offset(
        self, confusion_chart_data, url, project_id, model_id
    ):
        url_page1 = url + "?limit=10&offset=2"
        url_page2 = url + "?limit=10&offset=12"
        all_data = self.create_response_data(confusion_chart_data, 15)
        resp_data_1 = dict(count=10, previous=None, next=url_page2, data=all_data[2:12])
        resp_data_2 = dict(count=3, previous=url_page1, next=None, data=all_data[12:])
        responses.add(
            responses.GET, url_page1, json=resp_data_1, status=200, match_querystring=True
        )
        responses.add(
            responses.GET, url_page2, json=resp_data_2, status=200, match_querystring=True
        )

        confusion_charts = ExternalConfusionChart.list(project_id, model_id, limit=0, offset=2)
        assert len(confusion_charts) == 13
        assert confusion_charts[0].dataset_id == "dataset_id_2"

    @responses.activate
    def test_list_confusion_chart_limited(self, confusion_chart_data, url, project_id, model_id):
        url_page = url + "?limit=3&offset=2"
        all_data = self.create_response_data(confusion_chart_data, 5)
        resp_data = dict(count=3, previous=None, next=None, data=all_data[2:5])
        responses.add(responses.GET, url_page, json=resp_data, status=200, match_querystring=True)
        confusion_charts = ExternalConfusionChart.list(project_id, model_id, limit=3, offset=2)
        assert len(confusion_charts) == 3
        for i, chart in enumerate(confusion_charts):
            assert isinstance(chart, ExternalConfusionChart)
            self.assert_equal_confusion_chart(chart, all_data[i + 2])

    @responses.activate
    def test_get_confusion_chart_success(self, confusion_chart_data, url, project_id, model_id):
        dataset_id = "dataset_id_0"
        new_classes = ["4", "5", "6"]
        url_page = url + "{}/".format(dataset_id)
        metadata_url = url_page + "metadata/"
        all_data = self.create_response_data(confusion_chart_data, 2)
        metadata_response = {"classNames": new_classes}

        responses.add(responses.GET, metadata_url, json=metadata_response, status=200)
        responses.add(responses.GET, url_page, json=all_data[1], status=200, match_querystring=True)

        chart = ExternalConfusionChart.get(project_id, model_id, dataset_id)
        expected_chart_dict = all_data[1]
        expected_chart_dict["data"]["classes"] = ["4", "5", "6"]
        assert isinstance(chart, ExternalConfusionChart)
        self.assert_equal_confusion_chart(chart, expected_chart_dict)

    @responses.activate
    def test_get_confusion_chart_failure(self, url, project_id, model_id):
        dataset_id = "dataset_id_0"
        url_page = url + "{}/".format(dataset_id)

        exception_text = "No confusion chart for dataset {}".format(dataset_id)

        responses.add(responses.GET, url_page, json={"message": exception_text}, status=404)
        with pytest.raises(ClientError, match=exception_text):
            ExternalConfusionChart.get(project_id, model_id, dataset_id)



# -*- encoding: utf-8 -*-

"""This module is not considered part of the public interface. As of 2.3, anything here
may change or be removed without warning."""
from collections import defaultdict
from datetime import date, datetime
import re

from dateutil import parser, tz
import pandas as pd
import pytz
import six

from .deprecation import deprecated, deprecation_warning  # noqa
from .sourcedata import dataframe_to_buffer, is_urlsource, recognize_sourcedata  # noqa

ALL_CAPITAL = re.compile(r"(.)([A-Z][a-z]+)")
CASE_SWITCH = re.compile(r"([a-z0-9])([A-Z])")
UNDERSCORES = re.compile(r"([a-z]?)(_+)([a-z])")


class rawdict(dict):
    """
    Dictionaries returned from models will have their keys snakeCased.
    Wrapping them in rawdict will pass them through to the API verbatim.
    """

    pass


def underscorize(value):
    partial_result = ALL_CAPITAL.sub(r"\1_\2", value)
    return CASE_SWITCH.sub(r"\1_\2", partial_result).lower()


def underscoreToCamel(match):
    prefix, underscores, postfix = match.groups()
    if len(underscores) > 1:
        # underscoreToCamel('sample_pct__gte') -> 'samplePct__gte'
        return match.group()
    return prefix + postfix.upper()


def camelize(value):
    return UNDERSCORES.sub(underscoreToCamel, value)


def from_api(data, do_recursive=True, keep_attrs=None, keep_null_keys=False):
    if type(data) not in (dict, list):
        return data
    if isinstance(data, list):
        return _from_api_list(data, do_recursive=do_recursive, keep_null_keys=keep_null_keys)
    return _from_api_dict(
        data, do_recursive=do_recursive, keep_attrs=keep_attrs, keep_null_keys=keep_null_keys
    )


def _from_api_dict(data, do_recursive=True, keep_attrs=None, keep_null_keys=False):
    keep_attrs = keep_attrs or []
    # prepare attributes in format 'top.middle.bottom' for processing
    parsed_attrs = []
    for attr in keep_attrs:
        if isinstance(attr, six.string_types):
            parsed_attrs.append(attr.split("."))
        else:
            parsed_attrs.append(attr)
    # take index 0 since recursion goes from top to bottom
    current_level = [attr.pop(0) for attr in parsed_attrs if len(attr)]
    # filter out empty attrs to pass through recursive call
    next_level_attrs = [attr for attr in parsed_attrs if attr]

    app_data = {}
    for k, v in six.iteritems(data):
        k_under = underscorize(k)
        if v is None and k_under not in current_level and not keep_null_keys:
            continue
        if do_recursive:
            data_val = from_api(
                v,
                do_recursive=do_recursive,
                keep_attrs=next_level_attrs,
                keep_null_keys=keep_null_keys,
            )
        else:
            data_val = v
        app_data[k_under] = data_val
    return app_data


def _from_api_list(data, do_recursive=True, keep_null_keys=False):
    return [
        from_api(datum, do_recursive=do_recursive, keep_null_keys=keep_null_keys) for datum in data
    ]


def remove_empty_keys(metadata, keep_attrs=None):
    keep_attrs = keep_attrs or []
    return {k: v for k, v in metadata.items() if v is not None or k in keep_attrs}


def parse_time(time_str):
    try:
        return parser.parse(time_str, tzinfos={"UTC": tz.tzutc()})
    except Exception:
        return time_str


def datetime_to_string(datetime_obj, ensure_rfc_3339=False):
    """ Converts to isoformat
    """
    if not isinstance(datetime_obj, datetime):
        msg = "expected to be passed a datetime.datetime, was passed {}".format(type(datetime_obj))
        raise ValueError(msg)
    if ensure_rfc_3339 and not datetime_obj.tzinfo:
        datetime_obj = datetime_obj.replace(tzinfo=pytz.utc)
    return datetime_obj.isoformat()


def to_api(data, keep_attrs=None):
    """
    :param data: dictionary {'max_digits': 1}
    :return: {'maxDigits': 1}
    """
    if not data:
        return {}
    assert isinstance(data, dict), "Wrong type"
    return _to_api_item(data, keep_attrs)


def _to_api_item(item, keep_attrs=None):
    if isinstance(item, rawdict):
        return item
    elif isinstance(item, dict):
        dense_item = remove_empty_keys(item, keep_attrs)
        return {
            camelize(k): _to_api_item(v, keep_attrs=keep_attrs)
            for k, v in six.iteritems(dense_item)
        }
    elif isinstance(item, list):
        return [_to_api_item(subitem, keep_attrs=keep_attrs) for subitem in item]
    elif isinstance(item, datetime):
        return datetime_to_string(item)
    elif isinstance(item, date):
        return item.isoformat()
    else:
        return item


def get_id_from_response(response):
    location_string = response.headers["Location"]
    return get_id_from_location(location_string)


def get_id_from_location(location_string):
    return location_string.split("/")[-2]


def get_duplicate_features(features):
    duplicate_features = set()
    seen_features = set()
    for feature in features:
        if feature in seen_features:
            duplicate_features.add(feature)
        else:
            seen_features.add(feature)
    return list(duplicate_features)


def raw_prediction_response_to_dataframe(pred_response, class_prefix):
    """Raw predictions for classification come as nested json objects.

    This will extract it to be a single dataframe.

    Parameters
    ----------
    pred_response : dict
        The loaded json object returned from the prediction route.
    class_prefix : str
            The prefix to append to labels in the final dataframe (e.g., apple -> class_apple)

    Returns
    -------
    frame : pandas.DataFrame

    """
    response = from_api(pred_response)
    predictions = response["predictions"]
    task_type = response["task"]
    if task_type == "Multilabel":
        frame = _multilabel_predictions_to_dataframe(predictions, class_prefix)
    else:
        frame = pd.DataFrame.from_records(predictions)
        has_prediction_values = "prediction_values" in frame.columns
        if has_prediction_values:
            frame = _pivot_prediction_labels(frame, class_prefix)
        if "prediction_explanations" in frame.columns:
            frame = _pivot_prediction_explanations(frame, pred_response, has_prediction_values)
    return frame


def _multilabel_predictions_to_dataframe(predictions, class_prefix):
    """
    Converts the predictions from a json prediction API response and converts them into a DataFrame.

    Parameters
    ----------
    predictions : list(dict)
        The predictions as returned by the prediction API
    class_prefix : str
            The prefix to append to labels in the final dataframe (e.g., apple -> class_apple)

    Returns
    -------
    pandas.DataFrame
        The converted predictions
    """
    rows = []
    for prediction in predictions:
        predicted_labels = prediction["prediction"]
        row = {}
        row[u"row_id"] = prediction["row_id"]
        for prediction_value in prediction["prediction_values"]:
            label = prediction_value["label"]
            row[u"threshold_{}".format(label)] = prediction_value["threshold"]
            row[u"{}{}".format(class_prefix, label)] = prediction_value["value"]
            row[u"prediction_{}".format(label)] = 1 if label in predicted_labels else 0
        rows.append(row)
    return pd.DataFrame(rows)


def _pivot_prediction_explanations(frame, pred_response, has_prediction_values):
    """ Pivot the prediction_explanations from the dictionary to columns

    frame : DataFrame
        the DataFrame build based on predictions response
    pred_response : dict
        the response that contains explanations details applicable to all predictions
    has_prediction_values : bool
        flag showing whether there are prediction values for multiple class
    Returns
    -------
    frame : DataFrame
         the dataframe with parsed prediction explanations
    """
    _FEATURE = "Explanation_{}_feature_name"
    _FEATURE_VAL = "Explanation_{}_feature_value"
    _STRENGTH = "Explanation_{}_strength"

    columns = ()
    if has_prediction_values:
        columns += ("explained_class",)
    first_row = frame["prediction_explanations"][0]
    for i in range(len(first_row)):
        idx = i + 1
        columns += (
            _FEATURE.format(idx),
            _FEATURE_VAL.format(idx),
            _STRENGTH.format(idx),
        )
    data = []
    for prediction_explanations in frame["prediction_explanations"]:
        data_row = ()
        if has_prediction_values:
            data_row += (prediction_explanations[0]["label"],)
        for prediction_explanation in prediction_explanations:
            data_row += (
                prediction_explanation["feature"],
                prediction_explanation["feature_value"],
                prediction_explanation["strength"],
            )
        data.append(data_row)
    prediction_explanation_df = pd.DataFrame.from_records(data, columns=columns)
    frame = pd.concat([frame, prediction_explanation_df], axis=1)
    frame = frame.drop("prediction_explanations", axis=1)

    frame["shap_remaining_total"] = frame["prediction_explanation_metadata"].apply(
        lambda x: x.get("shap_remaining_total", None)
    )
    frame = frame.drop("prediction_explanation_metadata", axis=1)

    # shapBaseValue is either list with single float item or scalar float
    shap_base_value = pred_response["shapBaseValue"]
    if isinstance(shap_base_value, list):
        shap_base_value = shap_base_value[0]
    frame["shap_base_value"] = shap_base_value
    return frame


def _pivot_prediction_labels(frame, class_prefix):
    """ Pivot the prediciton_values from the dictionary to columns

    Parameters
    ----------
    frame : pandas.DataFrame

    Returns
    -------
    frame : pandas.DataFrame

    """
    wrapper = defaultdict(list)
    for pred_row in frame["prediction_values"]:
        for pred_value in pred_row:
            col_name = u"".join((class_prefix, u"{}".format(pred_value["label"])))
            wrapper[col_name].append(pred_value["value"])
    pred_frame = pd.DataFrame.from_records(wrapper)
    frame = pd.concat([frame, pred_frame], axis=1)
    return frame.drop("prediction_values", 1)


def encode_utf8_if_py2(string):
    """__repr__ is supposed to return string (bytes) in 2 and string (unicode) in 3
    this function can be used to convert our unicode to strings in Python 2 but leave them alone
    in Python 3"""
    return string.encode("utf-8") if six.PY2 else string



from copy import deepcopy
import json

import pytest
import responses

from datarobot import ExecutionEnvironmentVersion
from datarobot.enums import EXECUTION_ENVIRONMENT_VERSION_BUILD_STATUS
from datarobot.errors import AsyncTimeoutError
from tests.model.utils import assert_version


@pytest.fixture
def mocked_versions(mocked_execution_environment_versions):
    return mocked_execution_environment_versions


@pytest.fixture
def mocked_version(mocked_execution_environment_version):
    return mocked_execution_environment_version


@pytest.fixture
def make_versions_url(unittest_endpoint):
    def _make_versions_url(environment_id, version_id=None):
        base_url = "{}/executionEnvironments/{}/versions/".format(unittest_endpoint, environment_id)
        if version_id is not None:
            return "{}{}/".format(base_url, version_id)
        return base_url

    return _make_versions_url


def mock_get_response(url, response):
    responses.add(
        responses.GET, url, status=200, content_type="application/json", body=json.dumps(response),
    )


def test_from_server_data(mocked_version):
    version = ExecutionEnvironmentVersion.from_server_data(mocked_version)
    assert_version(version, mocked_version)


@responses.activate
def test_create_non_blocking(mocked_version, make_versions_url, tmpdir):
    responses.add(
        responses.POST,
        make_versions_url(mocked_version["environmentId"]),
        status=200,
        content_type="application/json",
        body=json.dumps({"id": mocked_version["id"]}),
    )
    responses.add(
        responses.GET,
        make_versions_url(mocked_version["environmentId"], mocked_version["id"]),
        status=200,
        content_type="application/json",
        body=json.dumps(mocked_version),
    )
    docker_context_file = tmpdir.mkdir("sub").join("docker_context")
    docker_context_file.write(b"docker_context archive content")
    docker_context_file_path = str(docker_context_file)

    version = ExecutionEnvironmentVersion.create(
        mocked_version["environmentId"],
        docker_context_file_path,
        mocked_version["label"],
        mocked_version["description"],
        max_wait=None,
    )
    assert_version(version, mocked_version)

    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == make_versions_url(mocked_version["environmentId"])
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == make_versions_url(
        mocked_version["environmentId"], mocked_version["id"]
    )


@responses.activate
def test_create_timeout(mocked_version, make_versions_url, tmpdir):
    responses.add(
        responses.POST,
        make_versions_url(mocked_version["environmentId"]),
        status=200,
        content_type="application/json",
        body=json.dumps({"id": mocked_version["id"]}),
    )

    mocked_version["buildStatus"] = "processing"
    mock_get_response(
        make_versions_url(mocked_version["environmentId"], mocked_version["id"]), mocked_version,
    )

    docker_context_file = tmpdir.mkdir("sub").join("docker_context")
    docker_context_file.write(b"docker_context archive content")
    docker_context_file_path = str(docker_context_file)

    with pytest.raises(AsyncTimeoutError):
        ExecutionEnvironmentVersion.create(
            mocked_version["environmentId"],
            docker_context_file_path,
            mocked_version["label"],
            mocked_version["description"],
            max_wait=0,
        )

    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == make_versions_url(mocked_version["environmentId"])
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == make_versions_url(
        mocked_version["environmentId"], mocked_version["id"]
    )


@responses.activate
@pytest.mark.parametrize("final_status", EXECUTION_ENVIRONMENT_VERSION_BUILD_STATUS.FINAL_STATUSES)
def test_create_blocking(mocked_version, make_versions_url, final_status, tmpdir):
    mocked_version1 = deepcopy(mocked_version)
    mocked_version2 = deepcopy(mocked_version)

    mocked_version1["buildStatus"] = "processing"
    mocked_version2["buildStatus"] = final_status

    responses.add(
        responses.POST,
        make_versions_url(mocked_version["environmentId"]),
        status=200,
        content_type="application/json",
        body=json.dumps({"id": mocked_version["id"]}),
    )

    get_url = make_versions_url(mocked_version["environmentId"], mocked_version["id"])
    mock_get_response(get_url, mocked_version1)
    mock_get_response(get_url, mocked_version2)

    docker_context_file = tmpdir.mkdir("sub").join("docker_context")
    docker_context_file.write(b"docker_context archive content")
    docker_context_file_path = str(docker_context_file)

    version = ExecutionEnvironmentVersion.create(
        mocked_version["environmentId"],
        docker_context_file_path,
        mocked_version["label"],
        mocked_version["description"],
    )
    assert_version(version, mocked_version2)

    assert len(responses.calls) == 3

    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == make_versions_url(mocked_version["environmentId"])
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == get_url
    assert responses.calls[2].request.method == "GET"
    assert responses.calls[2].request.url == get_url


@responses.activate
def test_get_version(mocked_version, make_versions_url):
    url = make_versions_url(mocked_version["environmentId"], mocked_version["id"])
    mock_get_response(url, mocked_version)

    version = ExecutionEnvironmentVersion.get(mocked_version["environmentId"], mocked_version["id"])
    assert_version(version, mocked_version)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_list_versions(mocked_versions, make_versions_url):
    environment_id = mocked_versions["data"][0]["environmentId"]
    url = make_versions_url(environment_id)
    mock_get_response(url, mocked_versions)

    versions = ExecutionEnvironmentVersion.list(environment_id)

    assert len(versions) == len(mocked_versions["data"])
    for version, mocked_version in zip(versions, mocked_versions["data"]):
        assert_version(version, mocked_version)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_list_versions_multiple_pages(mocked_versions, make_versions_url):
    environment_id = mocked_versions["data"][0]["environmentId"]

    url1 = make_versions_url(environment_id)
    url2 = make_versions_url(environment_id) + "2"

    mocked_versions_2nd = deepcopy(mocked_versions)
    mocked_versions["next"] = url2

    mock_get_response(url1, mocked_versions)
    mock_get_response(url2, mocked_versions_2nd)

    versions = ExecutionEnvironmentVersion.list(environment_id)
    assert len(versions) == len(mocked_versions["data"]) + len(mocked_versions_2nd["data"])
    for version, mocked_version in zip(
        versions, mocked_versions["data"] + mocked_versions_2nd["data"]
    ):
        assert_version(version, mocked_version)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url.endswith(url1)
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url.endswith(url2)


@responses.activate
def test_list_filter_build_status(mocked_version, make_versions_url):
    environment_id = mocked_version["environmentId"]
    url = make_versions_url(environment_id) + "?buildStatus={}".format(
        mocked_version["buildStatus"]
    )
    mock_get_response(url, {"count": 1, "next": None, "previous": None, "data": [mocked_version]})

    versions = ExecutionEnvironmentVersion.list(
        environment_id, EXECUTION_ENVIRONMENT_VERSION_BUILD_STATUS.SUCCESS
    )

    assert len(versions) == 1
    assert_version(versions[0], mocked_version)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
@pytest.mark.parametrize("expected_log", ["", "some log"])
@pytest.mark.parametrize("expected_err", ["", "some error"])
def test_get_build_log(mocked_version, make_versions_url, expected_log, expected_err):
    url = make_versions_url(mocked_version["environmentId"], mocked_version["id"])
    log_url = url + "buildLog/"

    mock_get_response(url, mocked_version)
    mock_get_response(log_url, {"log": expected_log, "error": expected_err})

    environment = ExecutionEnvironmentVersion.get(
        mocked_version["environmentId"], mocked_version["id"]
    )
    log, err = environment.get_build_log()

    assert log == expected_log
    assert err == (None if expected_err == "" else expected_err)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == log_url


@responses.activate
def test_refresh(mocked_version, make_versions_url):
    url = make_versions_url(mocked_version["environmentId"], mocked_version["id"])

    mock_get_response(url, mocked_version)

    mocked_version.update({"buildStatus": "failed", "label": "l"})

    mock_get_response(url, mocked_version)

    version = ExecutionEnvironmentVersion.get(mocked_version["environmentId"], mocked_version["id"])
    version.refresh()
    assert_version(version, mocked_version)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == url



import trafaret as t

from datarobot.enums import CUSTOM_MODEL_TARGET_TYPE, CUSTOM_TASK_TYPE
from datarobot.models.api_object import APIObject
from datarobot.models.custom_model_version import CustomModelVersion
from datarobot.utils import encode_utf8_if_py2
from datarobot.utils.pagination import unpaginate


class CustomTask(APIObject):
    """ A custom task.

    .. versionadded:: v2.25

    Attributes
    ----------
    id: str
        id of the custom task
    name: str
        name of the custom task
    language: str
        programming language of the custom task.
        Can be "python", "r", "java" or "other"
    description: str
        description of the custom task
    task_type: datarobot.CUSTOM_TASK_TYPE
        whether the task is a custom estimator or a custom transform.
        Values: [`datarobot.CUSTOM_TASK_TYPE.ESTIMATOR`, `datarobot.CUSTOM_TASK_TYPE.TRANSFORM`]
    target_type: datarobot.TARGET_TYPE or None
        target type if task is an estimator.
        Values: [`datarobot.TARGET_TYPE.BINARY`, `datarobot.TARGET_TYPE.REGRESSION`,
        `datarobot.TARGET_TYPE.MULTICLASS`, `datarobot.TARGET_TYPE.ANOMALY`]
    calibrate_predictions: bool
        whether anomaly predictions should be calibrated to be between 0 and 1 by DR.
        only applies to custom estimators with target type  `datarobot.TARGET_TYPE.ANOMALY`.
    latest_version: datarobot.CustomModelVersion or None
        latest version of the custom task if one exists
    deployments_count: int
        number of a deployments of the custom task
    created_by: str
        username of a user who user who created the custom task
    updated_at: str
        ISO-8601 formatted timestamp of when the custom task was updated
    created_at: str
        ISO-8601 formatted timestamp of when the custom task was created
    """

    _path = "customTasks/"

    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("task_type", optional=False, default=CUSTOM_TASK_TYPE.ESTIMATOR): t.String(),
            t.Key("target_type", optional=True, default=None): t.String(),
            t.Key("latest_version", optional=True, default=None): t.Or(
                CustomModelVersion.schema, t.Null()
            ),
            t.Key("created") >> "created_at": t.String(),
            t.Key("updated") >> "updated_at": t.String(),
            t.Key("name"): t.String(),
            t.Key("description"): t.String(allow_blank=True),
            t.Key("language", optional=True, default=None): t.String(),
            t.Key("created_by"): t.String(),
            t.Key("deployments_count"): t.Int(),
            t.Key("calibrate_predictions", optional=True, default=True): t.Bool(),
        }
    ).allow_extra("*")

    def __init__(self, **kwargs):
        self._set_values(**kwargs)

    def __repr__(self):
        return encode_utf8_if_py2(u"{}({!r})".format(self.__class__.__name__, self.name or self.id))

    def _set_values(
        self,
        id,
        name,
        description,
        latest_version,
        deployments_count,
        created_by,
        updated_at,
        created_at,
        target_type,
        task_type,
        language,
        calibrate_predictions,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.task_type = task_type
        self.target_type = target_type
        self.language = language
        self.calibrate_predictions = calibrate_predictions

        if self.task_type == CUSTOM_TASK_TYPE.ESTIMATOR:
            if not self.target_type:
                raise ValueError(
                    "Custom {} must have a target type".format(CUSTOM_TASK_TYPE.ESTIMATOR)
                )
        elif self.task_type == CUSTOM_TASK_TYPE.TRANSFORM:
            if self.target_type is not None:
                raise ValueError(
                    "Custom {} cannot have a target type".format(CUSTOM_TASK_TYPE.TRANSFORM)
                )
            else:
                self.target_type = self.task_type
        else:
            raise ValueError(
                "{} is not allowed task_type. Please select {} or {}".format(
                    self.task_type, CUSTOM_TASK_TYPE.TRANSFORM, CUSTOM_TASK_TYPE.ESTIMATOR
                )
            )

        self.latest_version = CustomModelVersion(**latest_version) if latest_version else None
        self.deployments_count = deployments_count
        self.created_by = created_by
        self.updated_at = updated_at
        self.created_at = created_at

    @classmethod
    def list(cls, is_deployed=None, order_by=None, search_for=None):
        """List custom tasks available to the user.

        .. versionadded:: v2.25

        Parameters
        ----------
        search_for: str, optional
            string for filtering custom tasks - only tasks that contain the
            string in name or description will be returned.
            If not specified, all custom task will be returned
        order_by: str, optional
            property to sort custom tasks by.
            Supported properties are "created" and "updated".
            Prefix the attribute name with a dash to sort in descending order,
            e.g. order_by='-created'.
            By default, the order_by parameter is None which will result in
            custom tasks being returned in order of creation time descending

        Returns
        -------
        List[CustomTask]
            a list of custom tasks.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        payload = {
            "order_by": order_by,
            "search_for": search_for,
        }
        data = unpaginate(cls._path, payload, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, custom_task_id):
        """Get custom task by id.

        .. versionadded:: v2.25

        Parameters
        ----------
        custom_task_id: str
            id of the custom task

        Returns
        -------
        CustomTask
            retrieved custom task

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = "{}{}/".format(cls._path, custom_task_id)
        data = cls._client.get(path).json()
        return cls.from_server_data(data)

    def download_latest_version(self, file_path):
        """Download the latest custom task version.

        .. versionadded:: v2.25

        Parameters
        ----------
        file_path: str
            path to create a file with custom task version content

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = "{}{}/download/".format(self._path, self.id)
        response = self._client.get(path)
        with open(file_path, "wb") as f:
            f.write(response.content)

    @classmethod
    def create(
        cls, name, task_type, target_type=None, description=None, calibrate_predictions=True,
    ):
        """Create a custom task.

        .. versionadded:: v2.25

        Parameters
        ----------
        name: str
            name of the custom task
        language: str
            programming language of the custom task.
            Can be "python", "r", "java" or "other"
        description: str, optional
            description of the custom task
        task_type: datarobot.CUSTOM_TASK_TYPE
            whether the task is a custom estimator or a custom transform.
            Values: [`datarobot.CUSTOM_TASK_TYPE.ESTIMATOR`, `datarobot.CUSTOM_TASK_TYPE.TRANSFORM`]
        target_type: datarobot.TARGET_TYPE or None
            target type if task is an estimator.
            should be None if task is a transform.
            Values: [`datarobot.TARGET_TYPE.BINARY`, `datarobot.TARGET_TYPE.REGRESSION`,
            `datarobot.TARGET_TYPE.MULTICLASS`, `datarobot.TARGET_TYPE.ANOMALY`]
        calibrate_predictions: bool, optional
            whether anomaly predictions should be calibrated to be between 0 and 1 by DR.
            defaults to True.
            only applies to custom estimators with target type  `datarobot.TARGET_TYPE.ANOMALY`.

        Returns
        -------
        CustomTask
            created a custom task

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        payload = dict(
            name=name, description=description, calibrate_predictions=calibrate_predictions,
        )
        if task_type == CUSTOM_TASK_TYPE.TRANSFORM:
            if target_type is not None:
                raise ValueError(
                    "Custom {} cannot have a target type".format(CUSTOM_TASK_TYPE.TRANSFORM)
                )
            else:
                payload["target_type"] = CUSTOM_TASK_TYPE.TRANSFORM
        elif task_type == CUSTOM_TASK_TYPE.ESTIMATOR:
            if target_type in CUSTOM_MODEL_TARGET_TYPE.TASK_TARGET_TYPES:
                payload["target_type"] = target_type
            elif target_type is None:
                raise ValueError(
                    "Custom {} must have a target type".format(CUSTOM_TASK_TYPE.ESTIMATOR)
                )
            else:
                raise ValueError("{} is not an allowed target type".format(target_type))
        else:
            raise ValueError(
                "{} is not allowed task_type. Please select {} or {}".format(
                    task_type, CUSTOM_TASK_TYPE.TRANSFORM, CUSTOM_TASK_TYPE.ESTIMATOR
                )
            )

        response = cls._client.post(cls._path, data=payload)
        return cls.from_server_data(response.json())

    @classmethod
    def copy_custom_task(
        cls, custom_task_id,
    ):
        """Create a custom task by copying existing one.

        .. versionadded:: v2.25

        Parameters
        ----------
        custom_task_id: str
            id of the custom task to copy

        Returns
        -------
        CustomTask
            created a custom task

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        path = "{}fromCustomTask/".format(cls._path)
        response = cls._client.post(path, data={"custom_task_id": custom_task_id})
        return cls.from_server_data(response.json())

    def update(self, name=None, description=None, **kwargs):
        """Update custom task properties.

        .. versionadded:: v2.25

        Parameters
        ----------
        name: str, optional
            new custom task name
        language: str, optional
            new custom task programming language
        description: str, optional
            new custom task description

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        payload = dict(name=name, description=description, **kwargs)
        url = "{}{}/".format(self._path, self.id)
        response = self._client.patch(url, data=payload)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))

    def refresh(self):
        """Update custom task with the latest data from server.

        .. versionadded:: v2.25

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = self._path.format(self.id)
        path = "{}{}/".format(url, self.id)

        response = self._client.get(path)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))

    def delete(self):
        """Delete custom task.

        .. versionadded:: v2.25

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = "{}{}/".format(self._path, self.id)
        self._client.delete(url)



from copy import deepcopy
import json
import os
import random

import dateutil
import mock
from pandas import DataFrame
import pytest
import responses
import six

from datarobot import errors
from datarobot.enums import DEFAULT_TIMEOUT
from datarobot.models.dataset import Dataset, DatasetDetails, FeatureTypeCount, ProjectLocation
from datarobot.utils import camelize, dataframe_to_buffer
from datarobot.utils.sourcedata import list_of_records_to_buffer
from tests.project.conftest import *  # noqa: E403
from tests.test_features import assert_dataset_feature
from tests.test_helpers import fixture_file_path
from tests.utils import request_body_to_json

BASIC_CREDS = {"credentialType": "basic", "user": "user123", "password": "pass123"}
S3_CREDS = {"credentialType": "s3", "awsAccessKeyId": "key123", "awsSecretAccessKey": "secret123"}
OAUTH_CREDS = {
    "credentialType": "oauth",
    "oauthRefreshToken": "token123",
    "oauthClientId": "client123",
    "oauthClientSecret": "secret123",
}
# GCP credentials are for allow extra params.
GCP_CREDS = {"credentialType": "gcp", "type": "aaa", "privateKeyId": "123", "privateKey": "456"}


@pytest.fixture
def file_path():
    file_path = "delme.csv"
    if os.path.exists(file_path):
        os.remove(file_path)

    yield file_path

    if os.path.exists(file_path):
        os.remove(file_path)


@pytest.fixture()
def create_dataset_response():
    return {
        "catalogVersionId": "5e31d20d5d5f030f3ea51b25",
        "statusId": "8dab34bc-9b71-406b-ae37-85626eec2584",
        "catalogId": "5e31d20c5d5f030f3ea51b24",
    }


@pytest.fixture()
def mock_dataset_details(dataset_id):
    return {
        "datasetId": dataset_id,
        "dataEngineQueryId": None,
        "dataSourceId": None,
        "dataSourceType": "local",
        "description": None,
        "eda1ModificationDate": "2020-01-09T19:11:11.372000Z",
        "eda1ModifierFullName": "Eric Shaw",
        "error": "",
        "featureCount": 31,
        "featureCountByType": [
            {"count": 2, "featureType": "Text"},
            {"count": 1, "featureType": "Boolean"},
            {"count": 18, "featureType": "Numeric"},
            {"count": 6, "featureType": "Categorical"},
        ],
        "lastModificationDate": "2020-01-09T19:11:11.856000Z",
        "lastModifierFullName": "Eric Shaw",
        "tags": [],
        "uri": "Anomaly_credit_card.csv",
    }


@pytest.fixture()
def empty_dataset_details(dataset_id):
    return {
        "datasetId": dataset_id,
        "dataEngineQueryId": None,
        "dataSourceId": None,
        "dataSourceType": "",
        "description": None,
        "eda1ModificationDate": None,
        "eda1ModifierFullName": None,
        "error": "",
        "featureCount": None,
        "featureCountByType": [],
        "lastModificationDate": "2020-02-18T22:17:05.586000Z",
        "lastModifierFullName": "Eric Shaw",
        "tags": [],
        "uri": "N/A",
    }


@pytest.fixture()
def empty_dataset():
    return {
        "isLatestVersion": True,
        "categories": [],
        "name": "Untitled Dataset",
        "datasetId": "5e39941679dd6b1d33332b42",
        "versionId": "5e39941679dd6b1d33332b43",
        "dataPersisted": None,
        "createdBy": "Eric Shaw",
        "creationDate": "2020-02-04T15:56:06.893000Z",
        "isSnapshot": False,
        "isDataEngineEligible": True,
        "datasetSize": 0,
        "rowCount": 0,
        "processingState": "RUNNING",
        "elysianFields": "so cool",
    }


@pytest.fixture()
def mock_dataset_full_response(mock_dataset, mock_dataset_details):
    response = mock_dataset_details
    response.update(mock_dataset)
    return response


@pytest.fixture()
def empty_dataset_full_response(empty_dataset, empty_dataset_details):
    response = empty_dataset_details
    response.update(empty_dataset)
    return response


@pytest.fixture()
def mock_dataset_list():
    return {
        "count": 2,
        "totalCount": 2,
        "next": None,
        "data": [
            {
                "isLatestVersion": True,
                "name": "Anomaly_credit_card.csv",
                "datasetId": "5e177c5bf6fe7c042641f118",
                "versionId": "5e177c5cf6fe7c042641f119",
                "categories": ["TRAINING", "PREDICTION"],
                "createdBy": "Eric Shaw",
                "dataPersisted": True,
                "creationDate": "2020-01-09T19:17:48.282000Z",
                "isSnapshot": True,
                "isDataEngineEligible": True,
                "datasetSize": 10993415,
                "rowCount": 26127,
                "processingState": "COMPLETED",
                "extraneousFields": ["electrical", "gravitational"],
            },
            {
                "isLatestVersion": True,
                "name": "Anomaly_credit_card.csv",
                "datasetId": "5e177ab185047702f27ef88a",
                "versionId": "5e177ab285047702f27ef88b",
                "categories": ["TRAINING", "PREDICTION"],
                "createdBy": "Eric Shaw",
                "dataPersisted": True,
                "creationDate": "2020-01-09T19:10:42.261000Z",
                "isSnapshot": True,
                "isDataEngineEligible": True,
                "datasetSize": 3456098,
                "rowCount": 32498,
                "processingState": "SOMETHING",
            },
        ],
    }


def assert_base_dataset_fields(api_obj, dataset_json):
    assert api_obj.version_id == dataset_json["versionId"]
    assert api_obj.name == dataset_json["name"]
    assert api_obj.categories == dataset_json["categories"]
    assert api_obj.created_at == dateutil.parser.parse(dataset_json["creationDate"])
    assert api_obj.created_by == dataset_json["createdBy"]
    assert api_obj.data_persisted == dataset_json["dataPersisted"]
    assert api_obj.is_data_engine_eligible == dataset_json["isDataEngineEligible"]
    assert api_obj.is_latest_version == dataset_json["isLatestVersion"]
    assert api_obj.is_snapshot == dataset_json["isSnapshot"]
    assert api_obj.size == dataset_json["datasetSize"]
    assert api_obj.row_count == dataset_json["rowCount"]
    assert api_obj.processing_state == dataset_json["processingState"]


def assert_dataset(dataset_obj, dataset_json):
    # type: (Dataset, dict) -> None
    assert dataset_obj.id == dataset_json["datasetId"]
    assert_base_dataset_fields(dataset_obj, dataset_json)


@pytest.fixture()
def dataset_get_response(dataset_url, mock_dataset_full_response):
    add_mock_get_response(dataset_url, mock_dataset_full_response)


@pytest.fixture()
def dataset_delete_response(dataset_url, mock_dataset):
    responses.add(
        responses.DELETE, dataset_url, status=204, content_type="application/json",
    )


def add_mock_get_response(url, json_dict):
    responses.add(
        responses.GET, url, status=200, content_type="application/json", body=json.dumps(json_dict)
    )


def test_from_server_data(mock_dataset_full_response):
    dataset = Dataset.from_server_data(mock_dataset_full_response)
    assert_dataset(dataset, mock_dataset_full_response)


def test_from_server_data_with_unknown_category_still_builds(mock_dataset):
    mock_dataset["categories"].append("foo")
    dataset = Dataset.from_server_data(mock_dataset)
    assert_dataset(dataset, mock_dataset)


def test_from_server_empty_dataset(empty_dataset_full_response):
    dataset = Dataset.from_server_data(empty_dataset_full_response)
    assert_dataset(dataset, empty_dataset_full_response)


@responses.activate
@pytest.mark.usefixtures("dataset_get_response")
def test_get_dataset(mock_dataset):
    dataset = Dataset.get(mock_dataset["datasetId"])
    assert_dataset(dataset, mock_dataset)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url.endswith("/datasets/{}/".format(dataset.id))


@responses.activate
@pytest.mark.usefixtures("dataset_get_response", "dataset_delete_response")
def test_delete_dataset(mock_dataset):
    dataset = Dataset.get(mock_dataset["datasetId"])
    Dataset.delete(dataset.id)

    assert responses.calls[1].request.method == "DELETE"
    assert responses.calls[1].request.url.endswith("/datasets/{}/".format(dataset.id))


@responses.activate
def test_un_delete_dataset(dataset_url, dataset_id):
    url = "{}deleted/".format(dataset_url)
    responses.add(
        responses.PATCH, url, status=204, content_type="application/json",
    )

    Dataset.un_delete(dataset_id)
    assert responses.calls[0].request.method == "PATCH"
    assert responses.calls[0].request.url.endswith("/datasets/{}/deleted/".format(dataset_id))


@responses.activate
def test_list_dataset(unittest_endpoint, mock_dataset_list):
    url = "{}/{}/".format(unittest_endpoint, "datasets")
    add_mock_get_response(url, mock_dataset_list)

    dataset_list = Dataset.list()
    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url.endswith("/datasets/")
    for dataset_obj, json_obj in zip(dataset_list, mock_dataset_list["data"]):
        assert_dataset(dataset_obj, json_obj)


@responses.activate
def test_list_dataset_multiple_pages(unittest_endpoint, mock_dataset_list):
    second_data_list = deepcopy(mock_dataset_list)
    next_link = "http://app.datarobot.com/api/datasets/2"
    mock_dataset_list["next"] = next_link
    url = "{}/{}/".format(unittest_endpoint, "datasets")

    add_mock_get_response(url, mock_dataset_list)
    add_mock_get_response(next_link, second_data_list)

    dataset_list = Dataset.list()
    assert len(dataset_list) == 2 * len(mock_dataset_list["data"])

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url.endswith("/datasets/")
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url.endswith(next_link)
    for dataset_obj, json_obj in zip(dataset_list, mock_dataset_list["data"] * 2):
        assert_dataset(dataset_obj, json_obj)


@responses.activate
def test_list_dataset_with_params(unittest_endpoint, mock_dataset_list):
    url = "{}/{}/".format(unittest_endpoint, "datasets")
    add_mock_get_response(url, mock_dataset_list)

    Dataset.list(category="thingy")
    assert get_url(0).endswith("?category=thingy")

    Dataset.list(order_by="thingies")
    assert get_url(1).endswith("?orderBy=thingies")

    Dataset.list(filter_failed=True)
    assert get_url(2).endswith("?filterFailed=true")


@responses.activate
def test_iterate_dataset_multiple_pages(unittest_endpoint, mock_dataset_list):
    second_data_list = deepcopy(mock_dataset_list)
    next_link = "http://app.datarobot.com/api/datasets/2"
    mock_dataset_list["next"] = next_link
    url = "{}/{}/".format(unittest_endpoint, "datasets")

    add_mock_get_response(url, mock_dataset_list)
    add_mock_get_response(next_link, second_data_list)

    dataset_iterator = Dataset.iterate()

    for el in mock_dataset_list["data"]:
        from_iterator = next(dataset_iterator)
        assert_dataset(from_iterator, el)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url.endswith("/datasets/")
    assert len(responses.calls) == 1

    for el in mock_dataset_list["data"]:
        from_iterator = next(dataset_iterator)
        assert_dataset(from_iterator, el)

    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url.endswith(next_link)

    assert list(dataset_iterator) == []


@responses.activate
def test_iterate_dataset_with_params(unittest_endpoint, mock_dataset_list):
    url = "{}/{}/".format(unittest_endpoint, "datasets")
    add_mock_get_response(url, mock_dataset_list)

    next(Dataset.iterate(category="thingy"))
    assert get_url(0).endswith("datasets/?category=thingy")

    next(Dataset.iterate(order_by="thingies"))
    assert get_url(1).endswith("datasets/?orderBy=thingies")

    next(Dataset.iterate(filter_failed=True))
    assert get_url(2).endswith("datasets/?filterFailed=true")

    next(Dataset.iterate(offset=5))
    assert get_url(3).endswith("datasets/?offset=5")

    next(Dataset.iterate(limit=3))
    assert get_url(4).endswith("datasets/?limit=3")


@responses.activate
def test_update(dataset_url, mock_dataset_obj, mock_dataset_list):
    updated_json = mock_dataset_list["data"][1]

    updated_json["datasetId"] = mock_dataset_obj.id

    with pytest.raises(AssertionError):
        assert_dataset(mock_dataset_obj, updated_json)

    add_mock_get_response(dataset_url, updated_json)
    mock_dataset_obj.update()
    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url.endswith("/datasets/{}/".format(mock_dataset_obj.id))
    assert_dataset(mock_dataset_obj, updated_json)


@responses.activate
def test_modify_name(dataset_url, mock_dataset):
    dataset = Dataset.from_server_data(mock_dataset)
    new_name = "new_name"
    mock_dataset["name"] = new_name
    assert dataset.name != new_name

    responses.add(
        responses.PATCH,
        dataset_url,
        status=200,
        content_type="application/json",
        body=json.dumps(mock_dataset),
    )
    dataset.modify(name=new_name)
    request = responses.calls[0].request
    request_body = request_body_to_json(request)
    assert request_body == {"name": new_name}
    assert request.method == "PATCH"
    assert request.url.endswith("/datasets/{}/".format(dataset.id))
    assert_dataset(dataset, mock_dataset)


@responses.activate
def test_modify_categories(dataset_url, mock_dataset):
    dataset = Dataset.from_server_data(mock_dataset)
    new_categories = []
    mock_dataset["categories"] = new_categories
    assert dataset.categories != new_categories

    responses.add(
        responses.PATCH,
        dataset_url,
        status=200,
        content_type="application/json",
        body=json.dumps(mock_dataset),
    )
    dataset.modify(categories=new_categories)
    request = responses.calls[0].request
    request_body = request_body_to_json(request)
    assert request_body == {"categories": new_categories}
    assert request.method == "PATCH"
    assert request.url.endswith("/datasets/{}/".format(dataset.id))
    assert_dataset(dataset, mock_dataset)


@responses.activate
def test_modify_with_no_params_makes_no_api_call_and_does_not_change_dataset(mock_dataset_obj):
    original_name = mock_dataset_obj.name
    original_categories = mock_dataset_obj.categories

    mock_dataset_obj.modify()

    assert len(responses.calls) == 0
    assert mock_dataset_obj.name == original_name
    assert mock_dataset_obj.categories == original_categories


@responses.activate
def test_get_projects(dataset_url, mock_dataset_obj):
    url = "{}projects/".format(dataset_url)
    next_link = "https://nextpage.com"
    location_one = {"url": "https://location.com/id-1/", "id": "id-1"}
    location_two = {"url": "https://location.com/id-2/", "id": "id-2"}
    projects_response_one = {
        "count": 1,
        "totalCount": 2,
        "next": next_link,
        "data": [location_one],
        "previous": None,
    }
    projects_response_two = {
        "count": 1,
        "totalCount": 2,
        "next": None,
        "data": [location_two],
        "previous": None,
    }
    add_mock_get_response(url, projects_response_one)
    add_mock_get_response(next_link, projects_response_two)

    result = mock_dataset_obj.get_projects()
    # noinspection PyArgumentList
    expected = {ProjectLocation(**location_one), ProjectLocation(**location_two)}
    assert set(result) == expected


class TestProjectCreate(object):
    @responses.activate
    @pytest.mark.usefixtures("project_creation_responses")
    def test_create_from_dataset_use_defaults(
        self, project_collection_url, project_without_target_json, mock_dataset_obj
    ):
        project = mock_dataset_obj.create_project()

        create_body = request_body_to_json(responses.calls[0].request)
        assert create_body == {
            "datasetId": mock_dataset_obj.id,
            "datasetVersionId": mock_dataset_obj.version_id,
        }
        assert responses.calls[0].request.url == project_collection_url

        response_json = json.loads(project_without_target_json)
        assert project.id == response_json["id"]
        assert project.project_name == response_json["projectName"]
        assert project.file_name == response_json["fileName"]
        assert project.stage == response_json["stage"]

    @responses.activate
    @pytest.mark.usefixtures("project_creation_responses")
    @pytest.mark.parametrize(
        "param_name,param_value",
        [
            ("project_name", "my cool name"),
            ("user", "that guy over there"),
            ("password", "my voice is my passport"),
            ("credential_id", "12345"),
            ("credential_data", BASIC_CREDS),
            ("credential_data", S3_CREDS),
            ("credential_data", OAUTH_CREDS),
            ("credential_data", GCP_CREDS),
            ("use_kerberos", True),
        ],
    )
    def test_create_from_dataset_using_params(self, param_name, param_value, mock_dataset_obj):
        mock_dataset_obj.create_project(**{param_name: param_value})
        create_body = request_body_to_json(responses.calls[0].request)
        assert create_body == {
            "datasetId": mock_dataset_obj.id,
            "datasetVersionId": mock_dataset_obj.version_id,
            camelize(param_name): param_value,
        }


class TestDatasetDownloadFile(object):
    @pytest.fixture()
    def downloaded_file(self):
        return b"a,b,c\n1,2,3\n4,5,6\n"

    @pytest.fixture
    def file_download_response(self, dataset_url, downloaded_file):
        url = "{}file/".format(dataset_url)
        responses.add(responses.GET, url, status=200, content_type="test/csv", body=downloaded_file)

    @responses.activate
    @pytest.mark.usefixtures("file_download_response")
    def test_get_file_with_file_path(self, mock_dataset_obj, downloaded_file, file_path):
        mock_dataset_obj.get_file(file_path=file_path)

        with open(file_path, "rb") as f:
            result = f.read()

        assert result == downloaded_file

    @responses.activate
    @pytest.mark.usefixtures("file_download_response")
    def test_get_file_with_filelike(self, mock_dataset_obj, downloaded_file, file_path):
        with open(file_path, "wb") as filelike:
            mock_dataset_obj.get_file(filelike=filelike)

        with open(file_path, "rb") as f:
            result = f.read()

        assert result == downloaded_file

    def test_get_file_with_neither_filelike_nor_file_path_raises_type_error(self, mock_dataset_obj):
        with pytest.raises(TypeError):
            mock_dataset_obj.get_file()

    def test_get_file_with_both_filelike_and_file_path_raises_type_error(
        self, mock_dataset_obj, file_path
    ):
        with pytest.raises(TypeError):
            with open(file_path, "wb") as filelike:
                mock_dataset_obj.get_file(file_path=file_path, filelike=filelike)


class TestDatasetGetFeatures(object):
    @pytest.fixture
    def feature_list_response(self):
        return {
            "count": 2,
            "totalCount": 52,
            "next": None,
            "data": [
                {
                    "timeSeriesEligibilityReason": "notADate",
                    "uniqueCount": 3,
                    "lowInformation": False,
                    "datasetVersionId": "5e1779e085047702f27ef884",
                    "name": "A1Cresult",
                    "dateFormat": None,
                    "max": 1,
                    "min": 2,
                    "median": 3,
                    "targetLeakage": "SKIPPED_DETECTION",
                    "featureType": "Numerical",
                    "naCount": 8379,
                    "stdDev": 4,
                    "timeStep": None,
                    "targetLeakageReason": "no target leakage was detected",
                    "timeSeriesEligible": False,
                    "timeUnit": None,
                    "id": 21,
                    "datasetId": "5e1779df85047702f27ef883",
                    "mean": 7,
                },
                {
                    "timeSeriesEligibilityReason": "notADate",
                    "uniqueCount": 30,
                    "lowInformation": False,
                    "datasetVersionId": "5e1779e085047702f27ef884",
                    "name": "acarbose",
                    "dateFormat": None,
                    "max": None,
                    "min": None,
                    "median": None,
                    "targetLeakage": "SKIPPED_DETECTION",
                    "featureType": "Categorical",
                    "naCount": 0,
                    "stdDev": None,
                    "timeStep": None,
                    "targetLeakageReason": "no target leakage was detected",
                    "timeSeriesEligible": False,
                    "timeUnit": None,
                    "id": 33,
                    "datasetId": "5e1779df85047702f27ef883",
                    "mean": None,
                },
            ],
            "previous": None,
        }

    @responses.activate
    def test_get_all_features(self, dataset_url, feature_list_response, mock_dataset_obj):
        url = "{}allFeaturesDetails/".format(dataset_url)

        add_mock_get_response(url, feature_list_response)
        result = mock_dataset_obj.get_all_features()

        json_list = feature_list_response["data"]
        assert len(result) == len(json_list)
        for feature_obj, feature_json in zip(result, json_list):
            assert_dataset_feature(feature_obj, feature_json)

        assert responses.calls[0].request.url == url

    @responses.activate
    def test_get_all_features_handles_pagination(
        self, dataset_url, feature_list_response, mock_dataset_obj
    ):
        url = "{}allFeaturesDetails/".format(dataset_url)
        second_page = "https://the-second-page.com/"

        first_page = deepcopy(feature_list_response)
        first_page["next"] = second_page

        add_mock_get_response(url, first_page)
        add_mock_get_response(second_page, feature_list_response)
        result = mock_dataset_obj.get_all_features()

        json_list = feature_list_response["data"] * 2
        assert len(result) == len(json_list)
        for feature_obj, feature_json in zip(result, json_list):
            assert_dataset_feature(feature_obj, feature_json)

        assert responses.calls[0].request.url == url
        assert responses.calls[1].request.url == second_page

    @responses.activate
    def test_get_all_features_order_by(self, dataset_url, mock_dataset_obj, feature_list_response):
        url = "{}allFeaturesDetails/".format(dataset_url)
        add_mock_get_response(url, feature_list_response)

        mock_dataset_obj.get_all_features(order_by="joe")
        assert responses.calls[0].request.url == "{}?orderBy={}".format(url, "joe")

    @responses.activate
    def test_iterate_all_features_handles_pagination(
        self, dataset_url, feature_list_response, mock_dataset_obj
    ):
        url = "{}allFeaturesDetails/".format(dataset_url)
        second_page = "https://the-second-page.com/"

        first_page = deepcopy(feature_list_response)
        first_page["next"] = second_page

        add_mock_get_response(url, first_page)
        add_mock_get_response(second_page, feature_list_response)

        features_iterator = mock_dataset_obj.iterate_all_features()

        for el in feature_list_response["data"]:
            from_iterator = next(features_iterator)
            assert_dataset_feature(from_iterator, el)

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url.endswith("/allFeaturesDetails/")
        assert len(responses.calls) == 1

        for el in feature_list_response["data"]:
            from_iterator = next(features_iterator)
            assert_dataset_feature(from_iterator, el)

        assert responses.calls[1].request.method == "GET"
        assert responses.calls[1].request.url.endswith(second_page)

        assert list(features_iterator) == []

    @responses.activate
    def test_iterate_all_features_with_params(
        self, dataset_url, feature_list_response, mock_dataset_obj
    ):
        url = "{}allFeaturesDetails/".format(dataset_url)
        add_mock_get_response(url, feature_list_response)

        next(mock_dataset_obj.iterate_all_features(order_by="thingies"))
        assert get_url(0).endswith("allFeaturesDetails/?orderBy=thingies")

        next(mock_dataset_obj.iterate_all_features(offset=5))
        assert get_url(1).endswith("allFeaturesDetails/?offset=5")

        next(mock_dataset_obj.iterate_all_features(limit=3))
        assert get_url(2).endswith("allFeaturesDetails/?limit=3")


class TestDatasetCreation(object):
    @pytest.fixture()
    def status_location(self):
        return "https://some_location.com/"

    @pytest.fixture()
    def set_status_response(self, dataset_url, mock_dataset, status_location):
        responses.add(responses.GET, status_location, status=303, headers={"Location": dataset_url})
        add_mock_get_response(url=dataset_url, json_dict=mock_dataset)
        return dataset_url

    @staticmethod
    def add_mock_upload_response(url, status_location):
        responses.add(responses.POST, url, status=200, adding_headers={"Location": status_location})

    @pytest.fixture()
    def set_up_file_upload(
        self, unittest_endpoint, status_location, set_status_response, file_path
    ):
        with open(file_path, "wb") as f:
            f.write(b"")
        self.add_mock_upload_response(
            "{}/datasets/fromFile/".format(unittest_endpoint), status_location
        )
        yield file_path

    @responses.activate
    def test_create_from_file_path(self, mock_dataset, set_up_file_upload, status_location):
        file_path = set_up_file_upload
        result = Dataset.create_from_file(file_path)

        post_call = responses.calls[0].request
        assert post_call.method == "POST"
        assert post_call.url.endswith("datasets/fromFile/")
        file_field = post_call.body.fields["file"]

        assert file_field[0] == os.path.basename(file_path)
        filelike = file_field[1]
        assert filelike.name == file_path

        status_call = responses.calls[1].request
        assert status_call.method == "GET"
        assert status_call.url == status_location

        get_dataset_call = responses.calls[2].request
        assert get_dataset_call.method == "GET"
        assert get_dataset_call.url.endswith("/datasets/{}/".format(mock_dataset["datasetId"]))

        assert_dataset(result, mock_dataset)

    @responses.activate
    def test_create_from_file_path_with_categories(
        self, mock_dataset, set_up_file_upload, dataset_url
    ):
        updated_dataset = deepcopy(mock_dataset)
        updated_dataset["categories"] = ["some", "new", "categories"]
        responses.add(
            responses.PATCH,
            dataset_url,
            status=200,
            content_type="application/json",
            body=json.dumps(updated_dataset),
        )
        categories = ["a", "b", "c"]
        result = Dataset.create_from_file(file_path=set_up_file_upload, categories=categories)
        patch_call = responses.calls[3].request
        assert patch_call.method == "PATCH"
        assert request_body_to_json(patch_call) == {"categories": categories}

        assert_dataset(result, updated_dataset)

    @responses.activate
    def test_create_from_filelike_with_name_field(self, mock_dataset, set_up_file_upload):
        file_path = set_up_file_upload
        with open(file_path, "wb") as file_pointer:
            # noinspection PyTypeChecker
            result = Dataset.create_from_file(filelike=file_pointer)

            post_call = responses.calls[0].request
            file_field = post_call.body.fields["file"]

            assert file_field[0] == os.path.basename(file_path)
            filelike = file_field[1]
            assert filelike.name == file_path

        assert_dataset(result, mock_dataset)

    @responses.activate
    def test_create_from_filelike_without_name_field(self, set_up_file_upload):
        file_contents = "hello"
        file_pointer = six.StringIO(file_contents)
        # noinspection PyTypeChecker
        Dataset.create_from_file(filelike=file_pointer)

        post_call = responses.calls[0].request
        file_field = post_call.body.fields["file"]

        assert file_field[0] == "data.csv"
        filelike = file_field[1]
        filelike.seek(0)
        assert filelike.read() == file_contents

    @responses.activate
    def test_create_from_file_uses_read_timeout(self):
        """
        Check that the read_timeout parameter gets passed correctly down into
        the `requests` library
        """
        path = fixture_file_path("synthetic-100.csv")
        read_timeout = 2

        # return 400 since all we care is that read_timeout is passed down to the HTTP request
        responses.add(
            responses.POST, "https://host_name.com/datasets/fromFile/", body="", status=400,
        )

        with mock.patch.object(
            Dataset._client, "request", wraps=Dataset._client.request
        ) as mock_request:
            try:
                Dataset.create_from_file(path, read_timeout=read_timeout)
            except errors.ClientError:
                pass

            timeout = mock_request.call_args[1]["timeout"]
            assert timeout[0] == DEFAULT_TIMEOUT.CONNECT  # Connect timeout;
            assert timeout[1] == read_timeout  # Read timeout;

    @responses.activate
    def test_create_from_file_uses_max_wait(self, mock_async_time, dataset_url):
        """
        Check that the max_wait parameter is in effect when uploading file
        """
        path = fixture_file_path("synthetic-100.csv")
        mock_async_time.time.side_effect = (0, 5)
        responses.add(
            responses.POST,
            "https://host_name.com/datasets/fromFile/",
            body="",
            status=200,
            headers={"Location": dataset_url},
        )
        responses.add(responses.GET, dataset_url, content_type="application/json", status=200)

        with pytest.raises(errors.AsyncTimeoutError):
            Dataset.create_from_file(path, max_wait=2)

    def test_create_from_file_no_file_params_raises_type_error(self):
        with pytest.raises(TypeError):
            Dataset.create_from_file(categories=["a", "b"])

    def test_create_from_file_two_file_params_raises_type_error(self):
        with pytest.raises(TypeError):
            # noinspection PyTypeChecker
            Dataset.create_from_file(file_path="a-path.csv", filelike=six.StringIO())

    @responses.activate
    def test_create_from_in_memory_data_with_dataframe(self, mock_dataset, set_up_file_upload):
        data = [{"a": "x", "b": "y"}, {"a": "d", "b": "e"}]
        df = DataFrame(data)
        test_buffer = dataframe_to_buffer(df)

        result = Dataset.create_from_in_memory_data(data_frame=df)

        post_call = responses.calls[0].request
        assert post_call.url.endswith("/datasets/fromFile/")
        assert post_call.method == "POST"
        file_field = post_call.body.fields["file"]
        assert file_field[0] == "data.csv"
        filelike = file_field[1]
        filelike.seek(0)
        assert filelike.read() == test_buffer.read()

        assert_dataset(result, mock_dataset)

    @responses.activate
    def test_create_from_in_memory_data_with_source_list(self, mock_dataset, set_up_file_upload):
        data = [{"a": "x", "b": "y"}, {"a": "d", "b": "e"}]
        test_buffer = list_of_records_to_buffer(data)

        result = Dataset.create_from_in_memory_data(records=data)

        post_call = responses.calls[0].request
        file_field = post_call.body.fields["file"]
        assert file_field[0] == "data.csv"
        filelike = file_field[1]
        filelike.seek(0)
        assert filelike.read() == test_buffer.read()

        assert_dataset(result, mock_dataset)

    @responses.activate
    def test_create_from_in_memory_data_with_categories(
        self, mock_dataset, set_up_file_upload, dataset_url
    ):
        updated_response = deepcopy(mock_dataset)
        updated_response["categories"] = ["some", "new", "categories"]
        responses.add(
            responses.PATCH,
            dataset_url,
            status=200,
            content_type="application/json",
            body=json.dumps(updated_response),
        )
        data = [{"a": "x", "b": "y"}, {"a": "d", "b": "e"}]
        categories = ["a", "b", "c"]
        result = Dataset.create_from_in_memory_data(records=data, categories=categories)

        patch_call = responses.calls[3].request
        assert patch_call.method == "PATCH"
        assert request_body_to_json(patch_call) == {"categories": categories}

        assert_dataset(result, updated_response)

    def test_create_from_in_memory_data_using_no_fields_raises_type_error(self):
        with pytest.raises(TypeError):
            Dataset.create_from_in_memory_data()

    def test_create_from_in_memory_data_using_both_fields_raises_type_error(self):
        data = [{"a": "x", "b": "y"}, {"a": "d", "b": "e"}]
        with pytest.raises(TypeError):
            Dataset.create_from_in_memory_data(data_frame=DataFrame(data), records=data)

    @pytest.fixture()
    def set_up_url_upload(self, unittest_endpoint, status_location, set_status_response):
        self.add_mock_upload_response(
            "{}/datasets/fromURL/".format(unittest_endpoint), status_location
        )

    @responses.activate
    @pytest.mark.usefixtures("set_up_url_upload")
    def test_create_from_url(self, mock_dataset):
        url = "https://somewhere.com"
        result = Dataset.create_from_url(url=url)

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {"url": url}

    @responses.activate
    @pytest.mark.usefixtures("set_up_url_upload")
    def test_create_from_url_with_do_snapshot(self, mock_dataset):
        do_snapshot = bool(random.randint(0, 1))
        url = "https://somewhere.com"
        result = Dataset.create_from_url(url=url, do_snapshot=do_snapshot)

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {"url": url, "doSnapshot": do_snapshot}

    @responses.activate
    @pytest.mark.usefixtures("set_up_url_upload")
    def test_create_from_url_with_persist_data_after_ingesiont(self, mock_dataset):
        persist_data_after_ingestion = bool(random.randint(0, 1))
        url = "https://somewhere.com"
        result = Dataset.create_from_url(
            url=url, persist_data_after_ingestion=persist_data_after_ingestion
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "url": url,
            "persistDataAfterIngestion": persist_data_after_ingestion,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_url_upload")
    def test_create_from_url_with_categories(self, mock_dataset):
        categories = ["a", "b", "c"]
        url = "https://somewhere.com"
        result = Dataset.create_from_url(url=url, categories=categories)

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {"url": url, "categories": categories}

    @pytest.fixture()
    def set_up_data_source_upload(self, unittest_endpoint, status_location, set_status_response):
        self.add_mock_upload_response(
            "{}/datasets/fromDataSource/".format(unittest_endpoint), status_location
        )

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    def test_create_from_data_source(self, mock_dataset):
        username = "test_user"
        password = "megasuperubersecretpassword"
        data_source_id = "5acc8437ec8d670001ba16bf"
        result = Dataset.create_from_data_source(
            data_source_id=data_source_id, username=username, password=password
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "user": username,
            "password": password,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    @pytest.mark.parametrize("do_snapshot", [True, False])
    def test_create_from_data_source_with_do_snapshot(self, mock_dataset, do_snapshot):
        username = "test_user"
        password = "megasuperubersecretpassword"
        data_source_id = "5acc8437ec8d670001ba16bf"
        result = Dataset.create_from_data_source(
            data_source_id=data_source_id,
            username=username,
            password=password,
            do_snapshot=do_snapshot,
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "user": username,
            "password": password,
            "doSnapshot": do_snapshot,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    @pytest.mark.parametrize("persist_data_after_ingestion", [True, False])
    def test_create_from_data_source_with_persist_data_after_ingestion(
        self, mock_dataset, persist_data_after_ingestion
    ):
        username = "test_user"
        password = "megasuperubersecretpassword"
        data_source_id = "5acc8437ec8d670001ba16bf"
        result = Dataset.create_from_data_source(
            data_source_id=data_source_id,
            username=username,
            password=password,
            persist_data_after_ingestion=persist_data_after_ingestion,
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "user": username,
            "password": password,
            "persistDataAfterIngestion": persist_data_after_ingestion,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    def test_create_from_data_source_with_categories(self, mock_dataset):
        categories = ["a", "b", "c"]
        username = "test_user"
        password = "megasuperubersecretpassword"
        data_source_id = "5acc8437ec8d670001ba16bf"
        result = Dataset.create_from_data_source(
            data_source_id=data_source_id,
            username=username,
            password=password,
            categories=categories,
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "user": username,
            "password": password,
            "categories": categories,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    @pytest.mark.parametrize("use_kerberos", [True, False])
    def test_create_from_data_source_with_use_kerberos(self, mock_dataset, use_kerberos):
        username = "test_user"
        password = "megasuperubersecretpassword"
        data_source_id = "5acc8437ec8d670001ba16bf"
        result = Dataset.create_from_data_source(
            data_source_id=data_source_id,
            username=username,
            password=password,
            use_kerberos=use_kerberos,
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "user": username,
            "password": password,
            "useKerberos": use_kerberos,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    def test_create_from_data_source_credentials(self, mock_dataset):
        data_source_id = "5acc8437ec8d670001ba16bf"
        credential_id = "6acc8437ec8d670001ba16bf"
        result = Dataset.create_from_data_source(
            data_source_id=data_source_id, credential_id=credential_id
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {"dataSourceId": data_source_id, "credentialId": credential_id}

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    @pytest.mark.parametrize("do_snapshot", [True, False])
    def test_create_from_data_source_with_do_snapshot_credentials(self, mock_dataset, do_snapshot):
        credential_id = "6acc8437ec8d670001ba16bf"
        data_source_id = "5acc8437ec8d670001ba16bf"
        result = Dataset.create_from_data_source(
            data_source_id=data_source_id, credential_id=credential_id, do_snapshot=do_snapshot
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "credentialId": credential_id,
            "doSnapshot": do_snapshot,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    @pytest.mark.parametrize("persist_data_after_ingestion", [True, False])
    def test_create_from_data_source_with_persist_data_after_ingestion_credentials(
        self, mock_dataset, persist_data_after_ingestion
    ):
        credential_id = "6acc8437ec8d670001ba16bf"
        data_source_id = "5acc8437ec8d670001ba16bf"
        result = Dataset.create_from_data_source(
            data_source_id=data_source_id,
            credential_id=credential_id,
            persist_data_after_ingestion=persist_data_after_ingestion,
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "credentialId": credential_id,
            "persistDataAfterIngestion": persist_data_after_ingestion,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    def test_create_from_data_source_with_categories_credentials(self, mock_dataset):
        categories = ["a", "b", "c"]
        credential_id = "6acc8437ec8d670001ba16bf"
        data_source_id = "5acc8437ec8d670001ba16bf"
        result = Dataset.create_from_data_source(
            data_source_id=data_source_id, credential_id=credential_id, categories=categories
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "credentialId": credential_id,
            "categories": categories,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    @pytest.mark.parametrize("use_kerberos", [True, False])
    def test_create_from_data_source_with_use_kerberos_credentials(
        self, mock_dataset, use_kerberos
    ):
        credential_id = "6acc8437ec8d670001ba16bf"
        data_source_id = "5acc8437ec8d670001ba16bf"
        result = Dataset.create_from_data_source(
            data_source_id=data_source_id, credential_id=credential_id, use_kerberos=use_kerberos
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "credentialId": credential_id,
            "useKerberos": use_kerberos,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    @pytest.mark.parametrize("credential_data", [BASIC_CREDS, S3_CREDS, OAUTH_CREDS, GCP_CREDS])
    def test_create_from_data_source_credential_data(self, mock_dataset, credential_data):
        data_source_id = "5acc8437ec8d670001ba16bf"

        result = Dataset.create_from_data_source(
            data_source_id=data_source_id, credential_data=credential_data
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {"dataSourceId": data_source_id, "credentialData": credential_data}

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    @pytest.mark.parametrize("do_snapshot", [True, False])
    @pytest.mark.parametrize("credential_data", [BASIC_CREDS, S3_CREDS, OAUTH_CREDS, GCP_CREDS])
    def test_create_from_data_source_with_do_snapshot_credential_data(
        self, mock_dataset, do_snapshot, credential_data
    ):
        data_source_id = "5acc8437ec8d670001ba16bf"

        result = Dataset.create_from_data_source(
            data_source_id=data_source_id, credential_data=credential_data, do_snapshot=do_snapshot
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "credentialData": credential_data,
            "doSnapshot": do_snapshot,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    @pytest.mark.parametrize("persist_data_after_ingestion", [True, False])
    @pytest.mark.parametrize("credential_data", [BASIC_CREDS, S3_CREDS, OAUTH_CREDS, GCP_CREDS])
    def test_create_from_data_source_with_persist_data_after_ingestion_credential_data(
        self, mock_dataset, persist_data_after_ingestion, credential_data
    ):
        data_source_id = "5acc8437ec8d670001ba16bf"

        result = Dataset.create_from_data_source(
            data_source_id=data_source_id,
            credential_data=credential_data,
            persist_data_after_ingestion=persist_data_after_ingestion,
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "credentialData": credential_data,
            "persistDataAfterIngestion": persist_data_after_ingestion,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    @pytest.mark.parametrize("credential_data", [BASIC_CREDS, S3_CREDS, OAUTH_CREDS, GCP_CREDS])
    def test_create_from_data_source_with_categories_credential_data(
        self, mock_dataset, credential_data
    ):
        categories = ["a", "b", "c"]
        data_source_id = "5acc8437ec8d670001ba16bf"

        result = Dataset.create_from_data_source(
            data_source_id=data_source_id, credential_data=credential_data, categories=categories
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "credentialData": credential_data,
            "categories": categories,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    @pytest.mark.parametrize("use_kerberos", [True, False])
    @pytest.mark.parametrize("credential_data", [BASIC_CREDS, S3_CREDS, OAUTH_CREDS, GCP_CREDS])
    def test_create_from_data_source_with_use_kerberos_credential_data(
        self, mock_dataset, use_kerberos, credential_data
    ):
        data_source_id = "5acc8437ec8d670001ba16bf"

        result = Dataset.create_from_data_source(
            data_source_id=data_source_id,
            credential_data=credential_data,
            use_kerberos=use_kerberos,
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "credentialData": credential_data,
            "useKerberos": use_kerberos,
        }


class TestDatasetDetails(object):
    @staticmethod
    def assert_details_equal(details_obj, details_json):
        # type: (DatasetDetails, dict) -> None

        def date_or_none(date_entry):
            return dateutil.parser.parse(date_entry) if date_entry else None

        assert details_obj.dataset_id == details_json["datasetId"]
        assert details_obj.data_source_type == details_json["dataSourceType"]
        assert details_obj.error == details_json["error"]
        assert details_obj.last_modification_date == dateutil.parser.parse(
            details_json["lastModificationDate"]
        )
        assert details_obj.last_modifier_full_name == details_json["lastModifierFullName"]
        assert details_obj.uri == details_json["uri"]
        assert details_obj.data_engine_query_id == details_json["dataEngineQueryId"]
        assert details_obj.data_source_id == details_json["dataSourceId"]
        assert details_obj.description == details_json["description"]
        assert details_obj.eda1_modification_date == date_or_none(
            details_json["eda1ModificationDate"]
        )
        assert details_obj.eda1_modifier_full_name == details_json["eda1ModifierFullName"]
        assert details_obj.feature_count == details_json["featureCount"]
        json_list = details_json["featureCountByType"]
        assert details_obj.feature_count_by_type == [
            FeatureTypeCount(feature_type=el["featureType"], count=el["count"]) for el in json_list
        ]
        assert details_obj.tags == details_json["tags"]

        assert_base_dataset_fields(details_obj, details_json)

    def test_create_from_server_data(self, mock_dataset_full_response):
        details = DatasetDetails.from_server_data(mock_dataset_full_response)
        self.assert_details_equal(details, mock_dataset_full_response)

    def test_create_from_server_data_using_empty_response(self, empty_dataset_full_response):
        details = DatasetDetails.from_server_data(empty_dataset_full_response)
        self.assert_details_equal(details, empty_dataset_full_response)

    @responses.activate
    def test_get_method(self, dataset_get_response, mock_dataset_full_response):
        details = DatasetDetails.get(mock_dataset_full_response["datasetId"])
        self.assert_details_equal(details, mock_dataset_full_response)

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url.endswith("/datasets/{}/".format(details.dataset_id))

    @responses.activate
    def test_to_dataset(self, dataset_get_response, mock_dataset_full_response):
        details = DatasetDetails.from_server_data(mock_dataset_full_response)
        dataset = details.to_dataset()
        assert details.dataset_id == dataset.id
        assert_dataset(dataset, mock_dataset_full_response)

        assert len(responses.calls) == 0

    @responses.activate
    def test_dataset_method_get_details(self, dataset_get_response, mock_dataset_full_response):
        dataset = Dataset.from_server_data(mock_dataset_full_response)
        details = dataset.get_details()

        assert details.dataset_id == dataset.id
        self.assert_details_equal(details, mock_dataset_full_response)

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url.endswith("/datasets/{}/".format(details.dataset_id))
        assert len(responses.calls) == 1


class TestDatasetVersionCreation(object):
    @pytest.fixture()
    def status_location(self):
        return "https://some_location.com/"

    @pytest.fixture()
    def set_status_response(self, dataset_url, mock_dataset, status_location):
        responses.add(responses.GET, status_location, status=303, headers={"Location": dataset_url})
        add_mock_get_response(url=dataset_url, json_dict=mock_dataset)
        return dataset_url

    @staticmethod
    def add_mock_upload_response(url, status_location):
        responses.add(responses.POST, url, status=200, adding_headers={"Location": status_location})

    @pytest.fixture()
    def set_up_file_upload(
        self, unittest_endpoint, status_location, set_status_response, file_path, mock_dataset
    ):
        with open(file_path, "wb") as f:
            f.write(b"")
        dataset_id = mock_dataset["datasetId"]
        self.add_mock_upload_response(
            "{}/datasets/{}/versions/fromFile/".format(unittest_endpoint, dataset_id),
            status_location,
        )
        yield file_path

    @responses.activate
    def test_create_from_file_path(self, mock_dataset, set_up_file_upload, status_location):
        file_path = set_up_file_upload
        result = Dataset.create_version_from_file(mock_dataset["datasetId"], file_path)

        post_call = responses.calls[0].request
        assert post_call.method == "POST"
        url = "datasets/{}/versions/fromFile/".format(mock_dataset["datasetId"])
        assert post_call.url.endswith(url)
        file_field = post_call.body.fields["file"]

        assert file_field[0] == os.path.basename(file_path)
        filelike = file_field[1]
        assert filelike.name == file_path

        status_call = responses.calls[1].request
        assert status_call.method == "GET"
        assert status_call.url == status_location

        get_dataset_call = responses.calls[2].request
        assert get_dataset_call.method == "GET"
        assert get_dataset_call.url.endswith("/datasets/{}/".format(mock_dataset["datasetId"]))

        assert_dataset(result, mock_dataset)

    @responses.activate
    def test_create_from_file_path_with_categories(
        self, mock_dataset, set_up_file_upload, dataset_url
    ):
        updated_dataset = deepcopy(mock_dataset)
        updated_dataset["categories"] = ["some", "new", "categories"]
        responses.add(
            responses.PATCH,
            dataset_url,
            status=200,
            content_type="application/json",
            body=json.dumps(updated_dataset),
        )
        categories = ["a", "b", "c"]
        result = Dataset.create_version_from_file(
            dataset_id=mock_dataset["datasetId"],
            file_path=set_up_file_upload,
            categories=categories,
        )
        patch_call = responses.calls[3].request
        assert patch_call.method == "PATCH"
        assert request_body_to_json(patch_call) == {"categories": categories}

        assert_dataset(result, updated_dataset)

    @responses.activate
    def test_create_from_filelike_with_name_field(self, mock_dataset, set_up_file_upload):
        file_path = set_up_file_upload
        with open(file_path, "wb") as file_pointer:
            # noinspection PyTypeChecker
            result = Dataset.create_version_from_file(
                dataset_id=mock_dataset["datasetId"], filelike=file_pointer
            )

            post_call = responses.calls[0].request
            file_field = post_call.body.fields["file"]

            assert file_field[0] == os.path.basename(file_path)
            filelike = file_field[1]
            assert filelike.name == file_path

        assert_dataset(result, mock_dataset)

    def test_create_from_file_no_file_params_raises_type_error(self, mock_dataset):
        with pytest.raises(TypeError):
            Dataset.create_version_from_file(
                dataset_id=mock_dataset["datasetId"], categories=["a", "b"]
            )

    def test_create_from_file_two_file_params_raises_type_error(self, mock_dataset):
        with pytest.raises(TypeError):
            # noinspection PyTypeChecker
            Dataset.create_version_from_file(
                dataset_id=mock_dataset["datasetId"],
                file_path="a-path.csv",
                filelike=six.StringIO(),
            )

    @responses.activate
    def test_create_from_in_memory_data_with_dataframe(self, mock_dataset, set_up_file_upload):
        data = [{"a": "x", "b": "y"}, {"a": "d", "b": "e"}]
        df = DataFrame(data)
        test_buffer = dataframe_to_buffer(df)

        result = Dataset.create_version_from_in_memory_data(
            dataset_id=mock_dataset["datasetId"], data_frame=df
        )

        post_call = responses.calls[0].request
        url = "datasets/{}/versions/fromFile/".format(mock_dataset["datasetId"])
        assert post_call.url.endswith(url)
        assert post_call.method == "POST"
        file_field = post_call.body.fields["file"]
        assert file_field[0] == "data.csv"
        filelike = file_field[1]
        filelike.seek(0)
        assert filelike.read() == test_buffer.read()

        assert_dataset(result, mock_dataset)

    @responses.activate
    def test_create_from_in_memory_data_with_source_list(self, mock_dataset, set_up_file_upload):
        data = [{"a": "x", "b": "y"}, {"a": "d", "b": "e"}]
        test_buffer = list_of_records_to_buffer(data)

        result = Dataset.create_version_from_in_memory_data(
            dataset_id=mock_dataset["datasetId"], records=data
        )

        post_call = responses.calls[0].request
        file_field = post_call.body.fields["file"]
        assert file_field[0] == "data.csv"
        filelike = file_field[1]
        filelike.seek(0)
        assert filelike.read() == test_buffer.read()

        assert_dataset(result, mock_dataset)

    def test_create_from_in_memory_data_using_no_fields_raises_type_error(self, mock_dataset):
        with pytest.raises(TypeError):
            Dataset.create_version_from_in_memory_data(dataset_id=mock_dataset["datasetId"])

    def test_create_from_in_memory_data_using_both_fields_raises_type_error(self, mock_dataset):
        data = [{"a": "x", "b": "y"}, {"a": "d", "b": "e"}]
        with pytest.raises(TypeError):
            Dataset.create_version_from_in_memory_data(
                dataset_id=mock_dataset["datasetId"], data_frame=DataFrame(data), records=data
            )

    @pytest.fixture()
    def set_up_url_upload(
        self, unittest_endpoint, status_location, set_status_response, mock_dataset
    ):
        self.add_mock_upload_response(
            "{}/datasets/{}/versions/fromURL/".format(unittest_endpoint, mock_dataset["datasetId"]),
            status_location,
        )

    @responses.activate
    @pytest.mark.usefixtures("set_up_url_upload")
    def test_create_from_url(self, mock_dataset):
        url = "https://somewhere.com"
        result = Dataset.create_version_from_url(dataset_id=mock_dataset["datasetId"], url=url)

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {"url": url}

    @responses.activate
    @pytest.mark.usefixtures("set_up_url_upload")
    def test_create_from_url_with_categories(self, mock_dataset):
        categories = ["a", "b", "c"]
        url = "https://somewhere.com"
        result = Dataset.create_version_from_url(
            dataset_id=mock_dataset["datasetId"], url=url, categories=categories
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {"url": url, "categories": categories}

    @pytest.fixture()
    def set_up_data_source_upload(
        self, unittest_endpoint, status_location, set_status_response, mock_dataset
    ):
        self.add_mock_upload_response(
            "{}/datasets/{}/versions/fromDataSource/".format(
                unittest_endpoint, mock_dataset["datasetId"]
            ),
            status_location,
        )

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    def test_create_from_data_source(self, mock_dataset):
        username = "test_user"
        password = "megasuperubersecretpassword"
        data_source_id = "5acc8437ec8d670001ba16bf"
        result = Dataset.create_version_from_data_source(
            dataset_id=mock_dataset["datasetId"],
            data_source_id=data_source_id,
            username=username,
            password=password,
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "user": username,
            "password": password,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    def test_create_from_data_source_with_categories(self, mock_dataset):
        categories = ["a", "b", "c"]
        username = "test_user"
        password = "megasuperubersecretpassword"
        data_source_id = "5acc8437ec8d670001ba16bf"
        result = Dataset.create_version_from_data_source(
            dataset_id=mock_dataset["datasetId"],
            data_source_id=data_source_id,
            username=username,
            password=password,
            categories=categories,
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "user": username,
            "password": password,
            "categories": categories,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    @pytest.mark.parametrize("use_kerberos", [True, False])
    def test_create_from_data_source_with_use_kerberos(self, mock_dataset, use_kerberos):
        username = "test_user"
        password = "megasuperubersecretpassword"
        data_source_id = "5acc8437ec8d670001ba16bf"
        result = Dataset.create_version_from_data_source(
            dataset_id=mock_dataset["datasetId"],
            data_source_id=data_source_id,
            username=username,
            password=password,
            use_kerberos=use_kerberos,
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "user": username,
            "password": password,
            "useKerberos": use_kerberos,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    def test_create_from_data_source_credentials(self, mock_dataset):
        data_source_id = "5acc8437ec8d670001ba16bf"
        credential_id = "6acc8437ec8d670001ba16bf"
        result = Dataset.create_version_from_data_source(
            dataset_id=mock_dataset["datasetId"],
            data_source_id=data_source_id,
            credential_id=credential_id,
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {"dataSourceId": data_source_id, "credentialId": credential_id}

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    def test_create_from_data_source_with_categories_credentials(self, mock_dataset):
        categories = ["a", "b", "c"]
        credential_id = "6acc8437ec8d670001ba16bf"
        data_source_id = "5acc8437ec8d670001ba16bf"
        result = Dataset.create_version_from_data_source(
            dataset_id=mock_dataset["datasetId"],
            data_source_id=data_source_id,
            credential_id=credential_id,
            categories=categories,
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "credentialId": credential_id,
            "categories": categories,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    @pytest.mark.parametrize("use_kerberos", [True, False])
    def test_create_from_data_source_with_use_kerberos_credentials(
        self, mock_dataset, use_kerberos
    ):
        credential_id = "6acc8437ec8d670001ba16bf"
        data_source_id = "5acc8437ec8d670001ba16bf"
        result = Dataset.create_version_from_data_source(
            dataset_id=mock_dataset["datasetId"],
            data_source_id=data_source_id,
            credential_id=credential_id,
            use_kerberos=use_kerberos,
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "credentialId": credential_id,
            "useKerberos": use_kerberos,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    @pytest.mark.parametrize("credential_data", [BASIC_CREDS, S3_CREDS, OAUTH_CREDS, GCP_CREDS])
    def test_create_from_data_source_credential_data(self, mock_dataset, credential_data):
        data_source_id = "5acc8437ec8d670001ba16bf"

        result = Dataset.create_version_from_data_source(
            dataset_id=mock_dataset["datasetId"],
            data_source_id=data_source_id,
            credential_data=credential_data,
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {"dataSourceId": data_source_id, "credentialData": credential_data}

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    @pytest.mark.parametrize("credential_data", [BASIC_CREDS, S3_CREDS, OAUTH_CREDS, GCP_CREDS])
    def test_create_from_data_source_with_categories_credential_data(
        self, mock_dataset, credential_data
    ):
        categories = ["a", "b", "c"]
        data_source_id = "5acc8437ec8d670001ba16bf"

        result = Dataset.create_version_from_data_source(
            dataset_id=mock_dataset["datasetId"],
            data_source_id=data_source_id,
            credential_data=credential_data,
            categories=categories,
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "credentialData": credential_data,
            "categories": categories,
        }

    @responses.activate
    @pytest.mark.usefixtures("set_up_data_source_upload")
    @pytest.mark.parametrize("use_kerberos", [True, False])
    @pytest.mark.parametrize("credential_data", [BASIC_CREDS, S3_CREDS, OAUTH_CREDS, GCP_CREDS])
    def test_create_from_data_source_with_use_kerberos_credential_data(
        self, mock_dataset, use_kerberos, credential_data
    ):
        data_source_id = "5acc8437ec8d670001ba16bf"

        result = Dataset.create_version_from_data_source(
            dataset_id=mock_dataset["datasetId"],
            data_source_id=data_source_id,
            credential_data=credential_data,
            use_kerberos=use_kerberos,
        )

        assert_dataset(result, mock_dataset)
        post_params = request_body_to_json(responses.calls[0].request)
        assert post_params == {
            "dataSourceId": data_source_id,
            "credentialData": credential_data,
            "useKerberos": use_kerberos,
        }


def get_url(index):
    return responses.calls[index].request.url
